from dataclasses import dataclass
import re
from io import StringIO
import json
from typing import Dict, List, Any
import yaml
try:
    from yaml import CLoader as Loader
except ImportError:
    from yaml import Loader


import pandas as pd
import requests
from requests.auth import HTTPBasicAuth

from mindsdb.integrations.utilities.sql_utils import (
    FilterCondition, FilterOperator, SortColumn
)
from mindsdb.integrations.libs.api_handler import APIResource


class ApiRequestException(Exception):
    pass


class ApiResponseException(Exception):
    pass


@dataclass
class APIInfo:
    """
    A class to store the information about the API.
    """
    base_url: str = None
    auth: dict = None


@dataclass
class APIEndpoint:
    url: str
    method: str
    params: dict
    response: dict


@dataclass
class APIResourceType:
    type_name: str
    sub_type: str = None
    properties: dict[str, str] = None


@dataclass
class APIEndpointParam:
    name: str
    type: APIResourceType
    where: str = None
    default: Any = None


def find_common_url_prefix(urls):
    if len(urls) == 0:
        return ''
    urls = [
        url.split('/')
        for url in urls
    ]

    min_len = min(len(s) for s in urls)

    for i in range(min_len):
        for j in range(1, len(urls)):
            if urls[j][i] != urls[0][i]:
                return '/'.join(urls[0][:i])

    return '/'.join(urls[0][:min_len])


class OpenAPISpecParser:
    """
    A class to parse the OpenAPI specification.
    """
    def __init__(self, openapi_spec_path: str) -> None:
        if openapi_spec_path.startswith('http://') or openapi_spec_path.startswith('https://'):
            response = requests.get(openapi_spec_path)
            response.raise_for_status()

            if openapi_spec_path.endswith('.json'):
                self.openapi_spec = response.json()
            else:
                stream = StringIO(response.text)
                self.openapi_spec = yaml.load(stream, Loader=Loader)
        else:
            raise ApiRequestException('URL is required')

    def get_security_schemes(self) -> dict:
        """
        Returns the security schemes defined in the OpenAPI specification.

        Returns:
            dict: A dictionary containing the security schemes defined in the OpenAPI specification.
        """
        return self.openapi_spec.get('components', {}).get('securitySchemes', {})

    def get_schemas(self) -> dict:
        """
        Returns the schemas defined in the OpenAPI specification.

        Returns:
            dict: A dictionary containing the schemas defined in the OpenAPI specification.
        """
        return self.openapi_spec.get('components', {}).get('schemas', {})

    def get_paths(self) -> dict:
        """
        Returns the paths defined in the OpenAPI specification.

        Returns:
            dict: A dictionary containing the paths defined in the OpenAPI specification.
        """
        return self.openapi_spec.get('paths', {})

    def get_specs(self) -> dict:
        return self.openapi_spec


class APIResourceGenerator:
    """
    A class to generate API resources based on the OpenAPI specification.
    """
    def __init__(self, url, connection_data, url_base=None, options=None) -> None:
        self.openapi_spec_parser = OpenAPISpecParser(url)
        self.connection_data = connection_data
        self.url_base = url_base
        self.options = options or {}
        self.resources = {}

    def check_connection(self):
        if 'check_connection_table' in self.options:
            table = self.resources.get(self.options['check_connection_table'])
            if table:
                table.list(targets=[], limit=1, conditions=[])

    def generate_api_resources(self, handler, table_name_format='{url}') -> Dict[str, APIResource]:
        """
        Generates an API resource based on the OpenAPI specification.

        Returns:
            Type[APIResource]: The generated API resource class.
        """
        paths = self.openapi_spec_parser.get_paths()
        schemas = self.openapi_spec_parser.get_schemas()
        self.security_schemes = self.openapi_spec_parser.get_security_schemes()

        self.resource_types = self.process_resource_types(schemas)
        endpoints = self.process_endpoints(paths)

        prefix_len = len(find_common_url_prefix([i.url for i in endpoints]))

        for endpoint in endpoints:
            url = endpoint.url[prefix_len:]
            # replace placehoders with x
            url = re.sub(r"{(\w+)}", 'x', url)
            url = url.replace('/', '_').strip('_')
            table_name = table_name_format.format(url=url, method=endpoint.method).lower()
            self.resources[table_name] = RestApiTable(handler, endpoint=endpoint, resource_gen=self)

        return self.resources

    def process_resource_types(self, schemas: dict) -> dict:
        resource_types = {}
        for name, schema_info in schemas.items():
            resource_types[name] = self._convert_to_resource_type(schema_info)

        return resource_types

    def process_endpoints(self, paths: dict) -> List[APIEndpoint]:
        """
        Processes the endpoints defined in the OpenAPI specification.

        Args:
            endpoints (Dict): A dictionary containing the endpoints defined in the OpenAPI specification.

        Returns:
            Dict: A dictionary containing the processed endpoints.
        """
        endpoints = []
        for path, path_info in paths.items():
            # filter endpoints by url base
            if self.url_base is not None and (not path.startswith(self.url_base) or path == self.url_base):
                continue

            for http_method, method_info in path_info.items():
                if http_method != 'get':
                    continue

                parameters = self._process_endpoint_parameters(method_info['parameters']) if 'parameters' in method_info else {}

                response = self._process_endpoint_response(method_info['responses'])
                if response['type'] is None:
                    continue

                endpoint = APIEndpoint(
                    url=path,
                    method=http_method,
                    params=parameters,
                    response=response
                )

                endpoints.append(endpoint)

        return endpoints

    def get_ref_object(self, ref):
        # get object by $ref link
        el = self.openapi_spec_parser.get_specs()
        for path in ref.lstrip('#').split('/'):
            if path:
                el = el[path]
        return el

    def _process_endpoint_parameters(self, parameters: list) -> Dict[str, APIEndpointParam]:
        """
        Processes the parameters defined in the OpenAPI specification.

        Args:
            parameters (Dict): A dictionary containing the parameters defined in the OpenAPI specification.

        Returns:
            Dict: A dictionary containing the processed parameters.
        """
        endpoint_parameters = {}
        for parameter in parameters:
            if '$ref' in parameter:
                parameter = self.get_ref_object(parameter['$ref'])

            type_name = self.get_resource_type(parameter['schema'])

            endpoint_parameters[parameter['name']] = APIEndpointParam(
                name=parameter['name'],
                type=type_name,
                default=parameter['schema'].get('default'),
                where=parameter['in'],
            )

        return endpoint_parameters

    def _process_endpoint_response(self, responses: dict):
        response = None
        response_path = []  # used to find list in response

        if '200' not in responses:
            return {'type': None}

        view = 'table'

        resp_success = responses['200']
        if '$ref' in resp_success:
            resp_success = self.get_ref_object(responses['200']['$ref'])

        for content_type, resp_info in resp_success['content'].items():
            if content_type != 'application/json':
                continue

            # type_name=get_type(resp_info['schema'])
            if 'schema' not in resp_info:
                continue

            resource_type = self._convert_to_resource_type(resp_info['schema'])

            # resolve type
            type_name = None
            if resource_type.type_name in self.resource_types:
                type_name = resource_type.type_name
                resource_type = self.resource_types[resource_type.type_name]

            if resource_type.type_name == 'array':
                response = resource_type.sub_type
            elif resource_type.type_name == 'object':
                if resource_type.properties is None:
                    raise NotImplementedError

                # if it is a table find property with list
                is_table = False
                if 'total_column' in self.options:
                    for col in self.options['total_column']:
                        if col in resource_type.properties:
                            is_table = True

                if is_table:
                    for k, v in resource_type.properties.items():
                        if v.type_name == 'array':

                            response = v.sub_type
                            response_path.append(k)
                            break
                else:
                    response = type_name
                    view = 'record'
            break

        return {
            'type': response,
            'path': response_path,
            'view': view
        }

    def _convert_to_resource_type(self, schema: dict) -> APIResourceType:
        """
        Converts the schema information to a resource type.

        Args:
            schema (Dict): A dictionary containing the schema information.

        Returns:
            APIResourceType: An object containing the resource type information.
        """
        type_name = self.get_resource_type(schema)
        # type_name= info['type']

        kwargs = {
            # 'name': name,
            'type_name': type_name,
        }

        if type_name == 'object':
            properties = {}
            if 'properties' in schema:
                for k, v in schema['properties'].items():
                    # type_name2 = get_type(v)
                    properties[k] = self._convert_to_resource_type(v)
            elif 'additionalProperties' in schema:
                if isinstance(schema['additionalProperties'], dict) and 'type' in schema['additionalProperties']:
                    type_name = schema['additionalProperties']['type']
                else:
                    type_name = 'string'

            kwargs['properties'] = properties
        if type_name == 'array' and 'items' in schema:
            kwargs['sub_type'] = self.get_resource_type(schema['items'])

        return APIResourceType(**kwargs)

    def get_resource_type(self, schema: dict) -> str:
        if 'type' in schema:
            return schema['type']

        elif '$ref' in schema:
            return schema['$ref'].split('/')[-1]

        elif 'allOf' in schema:
            # TODO Get only the first type.
            return self.get_resource_type(schema['allOf'][0])


class RestApiTable(APIResource):
    def __init__(self, *args, endpoint: APIEndpoint = None, resource_gen=None, **kwargs):
        self.endpoint = endpoint
        resource_types = resource_gen.resource_types
        self.connection_data = resource_gen.connection_data
        self.security_schemes = resource_gen.security_schemes
        self.options = resource_gen.options

        self.output_columns = {}
        response_type = endpoint.response['type']
        if response_type in resource_types:
            self.output_columns = resource_types[response_type].properties
        else:
            # let it be single column with this type
            self.output_columns = {'value': response_type}

        # check params:
        self.params, self.list_params = [], []
        for name, param in endpoint.params.items():
            self.params.append(name)
            if param.type == 'array':
                self.list_params.append(name)

        super().__init__(*args, **kwargs)

    def repr_value(self, value):
        # convert dict and lists to strings to show it response table

        if isinstance(value, dict):
            # remove empty keys
            value = {
                k: v
                for k, v in value.items()
                if v is not None
            }
            value = json.dumps(value)
        elif isinstance(value, list):
            value = ",".join([str(i) for i in value])
        return value

    def _handle_auth(self) -> dict:
        """
        Processes the authentication mechanism defined in the OpenAPI specification.
        Args:
            security_schemes (Dict): A dictionary containing the security schemes defined in the OpenAPI specification.
        Returns:
            Dict: A dictionary containing the authentication information required to connect to the API.
        """
        # API key authentication will be given preference over other mechanisms.
        # NOTE: If the API supports multiple authentication mechanisms, should they be supported? Which one should be given preference?

        security_schemes = self.security_schemes

        if 'token' in self.connection_data:
            headers = {'Authorization': f'Bearer {self.connection_data["token"]}'}

            return {
                "headers": headers
            }

        elif 'basicAuth' in security_schemes:
            # For basic authentication, the username and password are required.
            if not all(
                key in self.connection_data
                for key in ["username", "password"]
            ):
                raise ApiRequestException(
                    "The username and password are required for basic authentication."
                )
            return {
                "auth": HTTPBasicAuth(
                    self.connection_data["username"],
                    self.connection_data["password"],
                ),
            }
        return {}

    def get_columns(self) -> List[str]:
        return list(self.output_columns.keys())

    def get_setting_param(self, setting_name: str) -> str:
        # find input param name for specific setting

        if setting_name in self.options:
            for col in self.options[setting_name]:
                if col in self.endpoint.params:
                    return col

    def get_user_params(self):
        params = {}
        for k, v in self.connection_data.items():
            if k not in ('username', 'password', 'token', 'api_base'):
                params[k] = v
        return params

    def _api_request(self, filters):
        query, body, path_vars = {}, {}, {}
        for name, value in filters.items():
            param = self.endpoint.params[name]
            if param.where == 'query':
                query[name] = value
            elif param.where == 'path':
                path_vars[name] = value
            else:
                body[name] = value

        url = self.connection_data['api_base'] + self.endpoint.url
        if path_vars:
            url = url.format(**path_vars)
        # check empty placeholders
        placeholders = re.findall(r"{(\w+)}", url)
        if placeholders:
            raise ApiRequestException('Parameters are required: ' + ', '.join(placeholders))

        kwargs = self._handle_auth()
        req = requests.request(self.endpoint.method, url, params=query, data=body, **kwargs)

        if req.status_code != 200:
            raise ApiResponseException(req.text)
        resp = req.json()

        total = None
        if 'total_column' in self.options and isinstance(resp, dict):
            for col in self.options['total_column']:
                if col in resp:
                    total = resp[col]
                    break

        for item in self.endpoint.response['path']:
            resp = resp[item]

        if self.endpoint.response['view'] == 'record':
            # response is one record, make table
            resp = [resp]
        return resp, total

    def list(
        self,
        conditions: List[FilterCondition] = None,
        limit: int = None,
        sort: List[SortColumn] = None,
        targets: List[str] = None,
        **kwargs
    ) -> pd.DataFrame:

        if limit is None:
            limit = 20

        filters = {}
        if conditions:
            for condition in conditions:
                if condition.column not in self.params:
                    continue

                if condition.column in self.list_params:
                    if condition.op == FilterOperator.IN:
                        filters[condition.column] = condition.value
                    elif condition.op == FilterOperator.EQUAL:
                        filters[condition.column] = [condition]
                    condition.applied = True
                else:
                    filters[condition.column] = condition.value
                    condition.applied = True

        # user params
        params = self.get_user_params()
        if params:
            filters.update(params)

        page_size_param = self.get_setting_param('page_size_param')
        page_size = None
        if page_size_param is not None:
            # use default value for page size
            page_size = self.endpoint.params[page_size_param].default
            if page_size:
                filters[page_size_param] = page_size
        resp, total = self._api_request(filters)

        # pagination
        offset_param = self.get_setting_param('offset_param')
        page_num_param = self.get_setting_param('page_num_param')
        if offset_param is not None or page_num_param is not None:
            page_num = 1
            while True:
                count = len(resp)
                if limit <= count:
                    break

                if total is not None and total <= count:
                    # total is reached
                    break

                if page_size is not None and page_size > count:
                    # number of results are more than page, don't go to next page
                    break

                # download more pages
                if offset_param:
                    filters[offset_param] = count
                else:
                    page_num += 1
                    filters[page_num_param] = page_num
                resp2, total = self._api_request(filters)
                if len(resp2) == 0:
                    # no results from next page
                    break
                resp.extend(resp2)

        resp = resp[:limit]

        data = []

        columns = self.get_columns()
        for record in resp:
            item = {}

            if isinstance(record, dict):
                for name, value in record.items():
                    item[name] = self.repr_value(value)

                data.append(item)
            elif len(columns) > 0:
                # response is value
                item[columns[0]] = self.repr_value(record)

        return pd.DataFrame(data, columns=columns)



from collections import OrderedDict

from mindsdb.integrations.libs.const import HANDLER_CONNECTION_ARG_TYPE as ARG_TYPE


connection_args = OrderedDict(
    api_key={
        "type": ARG_TYPE.STR,
        "description": "API key",
        "required": True,
        "label": "api_key",
        "secret": True
    },
    sub_domain={
        "type": ARG_TYPE.STR,
        "description": "Sub-domain",
        "required": True,
        "label": "sub_domain",
        "secret": True
    },
    email={
        "type": ARG_TYPE.STR,
        "description": "Email ID",
        "required": True,
        "label": "email"
    }
)

connection_args_example = OrderedDict(
    api_key="api_key",
    sub_domain="sub_domain",
    email="email"
)



from mindsdb.integrations.libs.const import HANDLER_TYPE

from .__about__ import __version__ as version, __description__ as description
from .creation_args import creation_args
from .model_using_args import model_using_args
try:
    from .openai_handler import OpenAIHandler as Handler
    import_error = None
except Exception as e:
    Handler = None
    import_error = e

title = 'OpenAI'
name = 'openai'
type = HANDLER_TYPE.ML
icon_path = 'icon.svg'
permanent = False

__all__ = [
    'Handler', 'version', 'name', 'type', 'title', 'description',
    'import_error', 'icon_path', 'creation_args', 'model_using_args'
]



from typing import Any, List
import ast as py_ast

import pandas as pd
from mindsdb_sql_parser.ast import ASTNode, Select, Insert, Update, Delete, Star
from mindsdb_sql_parser.ast.select.identifier import Identifier

from mindsdb.integrations.utilities.sql_utils import (
    extract_comparison_conditions, filter_dataframe, sort_dataframe,
    FilterCondition, FilterOperator, SortColumn
)
from mindsdb.integrations.libs.base import BaseHandler
from mindsdb.integrations.libs.api_handler_exceptions import TableAlreadyExists, TableNotFound

from mindsdb.integrations.libs.response import (
    HandlerResponse as Response,
    RESPONSE_TYPE
)


class FuncParser:

    def from_string(self, query_string):

        body = py_ast.parse(query_string.strip(), mode='eval').body

        if not isinstance(body, py_ast.Call):
            raise RuntimeError(f'Api function not found {query_string}')

        fnc_name = body.func.id

        params = {}
        for keyword in body.keywords:
            name = keyword.arg
            value = self.process(keyword.value)

            params[name] = value

        return fnc_name, params

    def process(self, node):

        if isinstance(node, py_ast.List):
            elements = []
            for node2 in node.elts:
                elements.append(self.process(node2))
            return elements

        if isinstance(node, py_ast.Dict):

            keys = []
            for node2 in node.keys:
                if isinstance(node2, py_ast.Constant):
                    value = node2.value
                elif isinstance(node2, py_ast.Str):  # py37
                    value = node2.s
                else:
                    raise NotImplementedError(f'Unknown dict key {node2}')

                keys.append(value)

            values = []
            for node2 in node.values:
                values.append(self.process(node2))

            return dict(zip(keys, values))

        if isinstance(node, py_ast.Name):
            # special attributes
            name = node.id
            if name == 'true':
                return True
            elif name == 'false':
                return False
            elif name == 'null':
                return None

        if isinstance(node, py_ast.Constant):
            return node.value

        # ---- python 3.7 objects -----
        if isinstance(node, py_ast.Str):
            return node.s

        if isinstance(node, py_ast.Num):
            return node.n

        # -----------------------------

        if isinstance(node, py_ast.UnaryOp):
            if isinstance(node.op, py_ast.USub):
                value = self.process(node.operand)
                return -value

        raise NotImplementedError(f'Unknown node {node}')


class APITable:

    def __init__(self, handler):
        self.handler = handler

    def select(self, query: Select) -> pd.DataFrame:
        """Receive query as AST (abstract syntax tree) and act upon it.

        Args:
            query (ASTNode): sql query represented as AST. Usually it should be ast.Select

        Returns:
            pd.DataFrame
        """

        raise NotImplementedError()

    def insert(self, query: Insert) -> None:
        """Receive query as AST (abstract syntax tree) and act upon it somehow.

        Args:
            query (ASTNode): sql query represented as AST. Usually it should be ast.Insert

        Returns:
            None
        """
        raise NotImplementedError()

    def update(self, query: ASTNode) -> None:
        """Receive query as AST (abstract syntax tree) and act upon it somehow.

        Args:
            query (ASTNode): sql query represented as AST. Usually it should be ast.Update
        Returns:
            None
        """
        raise NotImplementedError()

    def delete(self, query: ASTNode) -> None:
        """Receive query as AST (abstract syntax tree) and act upon it somehow.

        Args:
            query (ASTNode): sql query represented as AST. Usually it should be ast.Delete

        Returns:
            None
        """
        raise NotImplementedError()

    def get_columns(self) -> list:
        """Maps the columns names from the API call resource

        Returns:
            List
        """
        raise NotImplementedError()


class APIResource(APITable):

    def __init__(self, *args, table_name=None, **kwargs):
        self.table_name = table_name
        super().__init__(*args, **kwargs)

    def select(self, query: Select) -> pd.DataFrame:
        """Receive query as AST (abstract syntax tree) and act upon it.

        Args:
            query (ASTNode): sql query represented as AST. Usually it should be ast.Select

        Returns:
            pd.DataFrame
        """

        conditions = self._extract_conditions(query.where)

        limit = None
        if query.limit:
            limit = query.limit.value

        sort = None
        if query.order_by and len(query.order_by) > 0:
            sort = []
            for an_order in query.order_by:
                sort.append(SortColumn(an_order.field.parts[-1],
                                       an_order.direction.upper() != 'DESC'))

        targets = []
        for col in query.targets:
            if isinstance(col, Identifier):
                targets.append(col.parts[-1])

        kwargs = {
            'conditions': conditions,
            'limit': limit,
            'sort': sort,
            'targets': targets
        }
        if self.table_name is not None:
            kwargs['table_name'] = self.table_name

        result = self.list(
            **kwargs
        )

        filters = []
        for cond in conditions:
            if not cond.applied:
                filters.append([cond.op.value, cond.column, cond.value])

        result = filter_dataframe(result, filters)

        if sort:
            sort_columns = []
            for idx, a_sort in enumerate(sort):
                if not a_sort.applied:
                    sort_columns.append(query.order_by[idx])

            result = sort_dataframe(result, sort_columns)

        if limit is not None and len(result) > limit:
            result = result[:int(limit)]

        return result

    def list(self,
             conditions: List[FilterCondition] = None,
             limit: int = None,
             sort: List[SortColumn] = None,
             targets: List[str] = None,
             **kwargs
             ):
        """
        List items based on specified conditions, limits, sorting, and targets.

        Args:
            conditions (List[FilterCondition]): Optional. A list of conditions to filter the items. Each condition
                                                should be an instance of the FilterCondition class.
            limit (int): Optional. An integer to limit the number of items to be listed.
            sort (List[SortColumn]): Optional. A list of sorting criteria
            targets (List[str]): Optional. A list of strings representing specific fields

        Raises:
            NotImplementedError: This is an abstract method and should be implemented in a subclass.
        """
        raise NotImplementedError()

    def insert(self, query: Insert) -> None:
        """Receive query as AST (abstract syntax tree) and act upon it somehow.

        Args:
            query (ASTNode): sql query represented as AST. Usually it should be ast.Insert

        Returns:
            None
        """

        columns = [col.name for col in query.columns]

        data = [
            dict(zip(columns, a_row))
            for a_row in query.values
        ]
        kwargs = {}
        if self.table_name is not None:
            kwargs['table_name'] = self.table_name

        self.add(data, **kwargs)

    def add(self, row: List[dict], **kwargs) -> None:
        """
        Add a single item to the dataa collection

        Args:
        r   ow (dict): A dictionary representing the item to be added.

        Raises:
            NotImplementedError: This is an abstract method and should be implemented in a subclass.
        """
        raise NotImplementedError()

    def update(self, query: Update) -> None:
        """Receive query as AST (abstract syntax tree) and act upon it somehow.

        Args:
            query (ASTNode): sql query represented as AST. Usually it should be ast.Update

        Returns:
            None
        """
        conditions = self._extract_conditions(query.where)

        values = {key: val.value for key, val in query.update_columns.items()}

        self.modify(conditions, values)

    def modify(self, conditions: List[FilterCondition], values: dict):
        """
        Modify items based on specified conditions and values.

        Args:
            conditions (List[FilterCondition]): A list of conditions to filter the items. Each condition
                                                should be an instance of the FilterCondition class.
            values (dict): A dictionary of values to be updated.

        Raises:
            NotImplementedError: This is an abstract method and should be implemented in a subclass.
        """
        raise NotImplementedError

    def delete(self, query: Delete) -> None:
        """Receive query as AST (abstract syntax tree) and act upon it somehow.

        Args:
            query (ASTNode): sql query represented as AST. Usually it should be ast.Delete

        Returns:
            None
        """
        conditions = self._extract_conditions(query.where)

        self.remove(conditions)

    def remove(self, conditions: List[FilterCondition]):
        """
        Remove items based on specified conditions.

        Args:
            conditions (List[FilterCondition]): A list of conditions to filter the items. Each condition
                                                should be an instance of the FilterCondition class.

        Raises:
            NotImplementedError: This is an abstract method and should be implemented in a subclass.
        """
        raise NotImplementedError()

    def _extract_conditions(self, where: ASTNode) -> List[FilterCondition]:
        return [
            FilterCondition(i[1], FilterOperator(i[0].upper()), i[2])
            for i in extract_comparison_conditions(where)
        ]


class APIHandler(BaseHandler):
    """
    Base class for handlers associated to the applications APIs (e.g. twitter, slack, discord  etc.)
    """

    def __init__(self, name: str):
        super().__init__(name)
        """ constructor
        Args:
            name (str): the handler name
        """

        self._tables = {}

    def _register_table(self, table_name: str, table_class: Any):
        """
        Register the data resource. For e.g if you are using Twitter API it registers the `tweets` resource from `/api/v2/tweets`.
        """
        if table_name in self._tables:
            raise TableAlreadyExists(f"Table with name {table_name} already exists for this handler")
        self._tables[table_name] = table_class

    def _get_table(self, name: Identifier):
        """
        Check if the table name was added to the _register_table
        Args:
            name (Identifier): the table name
        """
        name = name.parts[-1]
        if name not in self._tables:
            raise TableNotFound(f'Table not found: {name}')
        return self._tables[name]

    def query(self, query: ASTNode):

        if isinstance(query, Select):
            table = self._get_table(query.from_table)
            if not hasattr(table, 'list'):
                # for back compatibility, targets wasn't passed in previous version
                query.targets = [Star()]
            result = self._get_table(query.from_table).select(query)
        elif isinstance(query, Update):
            result = self._get_table(query.table).update(query)
        elif isinstance(query, Insert):
            result = self._get_table(query.table).insert(query)
        elif isinstance(query, Delete):
            result = self._get_table(query.table).delete(query)
        else:
            raise NotImplementedError

        if result is None:
            return Response(RESPONSE_TYPE.OK)
        elif isinstance(result, pd.DataFrame):
            return Response(RESPONSE_TYPE.TABLE, result)
        else:
            raise NotImplementedError

    def get_columns(self, table_name: str) -> Response:
        """
        Returns a list of entity columns
        Args:
            table_name (str): the table name
        Returns:
            RESPONSE_TYPE.TABLE
        """

        result = self._get_table(Identifier(table_name)).get_columns()

        df = pd.DataFrame(result, columns=['Field'])
        df['Type'] = 'str'

        return Response(RESPONSE_TYPE.TABLE, df)

    def get_tables(self) -> Response:
        """
        Return list of entities
        Returns:
            RESPONSE_TYPE.TABLE
        """
        result = list(self._tables.keys())

        df = pd.DataFrame(result, columns=['table_name'])
        df['table_type'] = 'BASE TABLE'

        return Response(RESPONSE_TYPE.TABLE, df)


class APIChatHandler(APIHandler):

    def get_chat_config(self):
        """Return configuration to connect to chatbot

        Returns:
            Dict
        """
        raise NotImplementedError()

    def get_my_user_name(self) -> list:
        """Return configuration to connect to chatbot

        Returns:
            Dict
        """
        raise NotImplementedError()



import pandas as pd
from typing import List
from mindsdb.integrations.libs.api_handler import APITable
from mindsdb.integrations.utilities.handlers.query_utilities import SELECTQueryParser, SELECTQueryExecutor
from mindsdb.utilities import log
from mindsdb_sql_parser import ast

logger = log.getLogger(__name__)


class DockerHubRepoImagesSummaryTable(APITable):
    """The DockerHub Repo Images Summary Table implementation"""

    def select(self, query: ast.Select) -> pd.DataFrame:
        """Pulls data from the https://docs.docker.com/docker-hub/api/latest/#tag/images" API

        Parameters
        ----------
        query : ast.Select
           Given SQL SELECT query

        Returns
        -------
        pd.DataFrame
            repo images summary matching the query

        Raises
        ------
        ValueError
            If the query contains an unsupported condition
        """

        select_statement_parser = SELECTQueryParser(
            query,
            'repo_images_summary',
            self.get_columns()
        )

        selected_columns, where_conditions, order_by_conditions, result_limit = select_statement_parser.parse_query()

        search_params = {}
        subset_where_conditions = []
        for op, arg1, arg2 in where_conditions:
            if arg1 == 'namespace':
                if op == '=':
                    search_params["namespace"] = arg2
                else:
                    raise NotImplementedError("Only '=' operator is supported for namespace column.")
            elif arg1 == 'repository':
                if op == '=':
                    search_params["repository"] = arg2
                else:
                    raise NotImplementedError("Only '=' operator is supported for repository column.")
            elif arg1 in self.get_columns():
                subset_where_conditions.append([op, arg1, arg2])

        filter_flag = ("namespace" in search_params) and ("repository" in search_params)

        if not filter_flag:
            raise NotImplementedError("Both namespace and repository columns have to be present in WHERE clause.")

        repo_images_summary_df = pd.DataFrame(columns=self.get_columns())

        response = self.handler.docker_client.get_images_summary(search_params["namespace"], search_params["repository"])

        self.check_res(res=response)

        content = response["content"]

        repo_images_summary_df = pd.json_normalize({"active_from": content["active_from"], "total": content["statistics"]["total"], "active": content["statistics"]["active"], "inactive": content["statistics"]["inactive"]})

        select_statement_executor = SELECTQueryExecutor(
            repo_images_summary_df,
            selected_columns,
            subset_where_conditions,
            order_by_conditions,
            result_limit
        )

        repo_images_summary_df = select_statement_executor.execute_query()

        return repo_images_summary_df

    def check_res(self, res):
        if res["code"] != 200:
            raise Exception("Error fetching results - " + res["error"])

    def get_columns(self) -> List[str]:
        """Gets all columns to be returned in pandas DataFrame responses

        Returns
        -------
        List[str]
            List of columns
        """

        return [
            "active_from",
            "total",
            "active",
            "inactive"
        ]
      
class DockerHubOrgSettingsTable(APITable):
    """The DockerHub Repo Org Settings Table implementation"""

    def select(self, query: ast.Select) -> pd.DataFrame:
        """Pulls data from the https://hub.docker.com/v2/orgs/{name}/settings" API

        Parameters
        ----------
        query : ast.Select
           Given SQL SELECT query

        Returns
        -------
        pd.DataFrame
            org settings matching the query

        Raises
        ------
        ValueError
            If the query contains an unsupported condition
        """

        select_statement_parser = SELECTQueryParser(
            query,
            'org_settings',
            self.get_columns()
        )

        selected_columns, where_conditions, order_by_conditions, result_limit = select_statement_parser.parse_query()

        search_params = {}
        subset_where_conditions = []
        for op, arg1, arg2 in where_conditions:
            if arg1 == 'organization':
                if op == '=':
                    search_params["organization"] = arg2
                else:
                    raise NotImplementedError("Only '=' operator is supported for organization column.")
            elif arg1 in self.get_columns():
                subset_where_conditions.append([op, arg1, arg2])

        if "organization" not in search_params:
            raise NotImplementedError("organization column has to be present in where clause.")

        organization_df = pd.DataFrame(columns=self.get_columns())

        response = self.handler.docker_client.get_org_settings(search_params["organization"])

        self.check_res(res=response)

        content = response["content"]

        organization_df = pd.json_normalize({"restricted_images_enabled": content["restricted_images"]["enabled"], "restricted_images_allow_official_images": content["restricted_images"]["allow_official_images"], "restricted_images_allow_verified_publishers": content["restricted_images"]["allow_verified_publishers"]})

        select_statement_executor = SELECTQueryExecutor(
            organization_df,
            selected_columns,
            subset_where_conditions,
            order_by_conditions,
            result_limit
        )

        organization_df = select_statement_executor.execute_query()

        return organization_df

    def check_res(self, res):
        if res["code"] != 200:
            raise Exception("Error fetching results - " + res["error"])

    def get_columns(self) -> List[str]:
        """Gets all columns to be returned in pandas DataFrame responses

        Returns
        -------
        List[str]
            List of columns
        """

        return [
            "restricted_images_enabled",
            "restricted_images_allow_official_images",
            "restricted_images_allow_verified_publishers"
        ]

class DockerHubRepoImagesTable(APITable):
    """The DockerHub Repo Images Table implementation"""

    def select(self, query: ast.Select) -> pd.DataFrame:
        """Pulls data from the https://hub.docker.com/v2/namespaces/{namespace}/repositories/{repository}/images" API

        Parameters
        ----------
        query : ast.Select
           Given SQL SELECT query

        Returns
        -------
        pd.DataFrame
            Repo Images matching the query

        Raises
        ------
        ValueError
            If the query contains an unsupported condition
        """

        select_statement_parser = SELECTQueryParser(
            query,
            'repo_images',
            self.get_columns()
        )

        selected_columns, where_conditions, order_by_conditions, result_limit = select_statement_parser.parse_query()

        search_params = {}
        subset_where_conditions = []
        for op, arg1, arg2 in where_conditions:
            if arg1 == 'namespace':
                if op == '=':
                    search_params["namespace"] = arg2
                else:
                    raise NotImplementedError("Only '=' operator is supported for namespace column.")
            elif arg1 == 'repository':
                if op == '=':
                    search_params["repository"] = arg2
                else:
                    raise NotImplementedError("Only '=' operator is supported for repository column.")
            elif arg1 in self.get_columns():
                subset_where_conditions.append([op, arg1, arg2])

        filter_flag = ("namespace" in search_params) and ("repository" in search_params)

        if not filter_flag:
            raise NotImplementedError("namespace and repository column has to be present in where clause.")

        repo_images_summary_df = pd.DataFrame(columns=self.get_columns())

        response = self.handler.docker_client.get_repo_images(search_params["namespace"], search_params["repository"])

        self.check_res(res=response)

        content = response["content"]

        repo_images_summary_df = pd.json_normalize(content["results"])

        select_statement_executor = SELECTQueryExecutor(
            repo_images_summary_df,
            selected_columns,
            subset_where_conditions,
            order_by_conditions,
            result_limit
        )

        repo_images_summary_df = select_statement_executor.execute_query()

        return repo_images_summary_df

    def check_res(self, res):
        if res["code"] != 200:
            raise Exception("Error fetching results - " + res["error"])

    def get_columns(self) -> List[str]:
        """Gets all columns to be returned in pandas DataFrame responses

        Returns
        -------
        List[str]
            List of columns
        """

        return ["namespace",
            "repository",
            "digest",
            "tags",
            "last_pushed",
            "last_pulled",
            "status"
            ]


class DockerHubRepoTagTable(APITable):
    """The DockerHub Repo Tag Table implementation"""

    def select(self, query: ast.Select) -> pd.DataFrame:
        """Pulls data from the https://docs.docker.com/docker-hub/api/latest/#tag/images" API

        Parameters
        ----------
        query : ast.Select
           Given SQL SELECT query

        Returns
        -------
        pd.DataFrame
            Repo Tag matching the query

        Raises
        ------
        ValueError
            If the query contains an unsupported condition
        """

        select_statement_parser = SELECTQueryParser(
            query,
            'repo_tag_details',
            self.get_columns()
        )

        selected_columns, where_conditions, order_by_conditions, result_limit = select_statement_parser.parse_query()

        search_params = {}
        subset_where_conditions = []
        for op, arg1, arg2 in where_conditions:
            if arg1 == 'namespace':
                if op == '=':
                    search_params["namespace"] = arg2
                else:
                    raise NotImplementedError("Only '=' operator is supported for namespace column.")
            elif arg1 == 'repository':
                if op == '=':
                    search_params["repository"] = arg2
                else:
                    raise NotImplementedError("Only '=' operator is supported for repository column.")
            elif arg1 == 'tag':
                if op == '=':
                    search_params["tag"] = arg2
                else:
                    raise NotImplementedError("Only '=' operator is supported for tag column.")
            elif arg1 in self.get_columns():
                subset_where_conditions.append([op, arg1, arg2])

        filter_flag = ("namespace" in search_params) and ("repository" in search_params) and ("tag" in search_params)

        if not filter_flag:
            raise NotImplementedError("namespace, repository and tag column has to be present in where clause.")

        repo_tag_summary_df = pd.DataFrame(columns=self.get_columns())

        response = self.handler.docker_client.get_repo_tag(search_params["namespace"], search_params["repository"], search_params["tag"])

        self.check_res(res=response)

        content = response["content"]

        repo_tag_summary_df = pd.json_normalize({"creator": content["creator"], 
            "id": content["id"], 
            "images": content["images"], 
            "last_updated": content["last_updated"], 
            "last_updater": content["last_updater"], 
            "last_updater_username": content["last_updater_username"], 
            "name": content["name"], 
            "repository": content["repository"], 
            "full_size": content["full_size"], 
            "v2": content["v2"], 
            "tag_status": content["tag_status"], 
            "tag_last_pulled": content["tag_last_pulled"], 
            "tag_last_pushed": content["tag_last_pushed"], 
            "media_type": content["media_type"], 
            "content_type": content["media_type"]
        })

        select_statement_executor = SELECTQueryExecutor(
            repo_tag_summary_df,
            selected_columns,
            subset_where_conditions,
            order_by_conditions,
            result_limit
        )

        repo_tag_summary_df = select_statement_executor.execute_query()

        return repo_tag_summary_df

    def check_res(self, res):
        if res["code"] != 200:
            raise Exception("Error fetching results - " + res["error"])

    def get_columns(self) -> List[str]:
        """Gets all columns to be returned in pandas DataFrame responses

        Returns
        -------
        List[str]
            List of columns
        """

        return ["creator", 
                "id", 
                "images", 
                "last_updated", 
                "last_updater", 
                "last_updater_username", 
                "name", 
                "repository", 
                "full_size", 
                "v2", 
                "tag_status", 
                "tag_last_pulled", 
                "tag_last_pushed", 
                "media_type", 
                "content_type"
            ]

class DockerHubRepoTagsTable(APITable):
    """The DockerHub Repo Tags Table implementation"""

    def select(self, query: ast.Select) -> pd.DataFrame:
        """Pulls data from the https://hub.docker.com/v2/namespaces/{namespace}/repositories/{repository}/tags" API

        Parameters
        ----------
        query : ast.Select
           Given SQL SELECT query

        Returns
        -------
        pd.DataFrame
            Repo Tag matching the query

        Raises
        ------
        ValueError
            If the query contains an unsupported condition
        """

        select_statement_parser = SELECTQueryParser(
            query,
            'repo_tags',
            self.get_columns()
        )

        selected_columns, where_conditions, order_by_conditions, result_limit = select_statement_parser.parse_query()

        search_params = {}
        subset_where_conditions = []
        for op, arg1, arg2 in where_conditions:
            if arg1 == 'namespace':
                if op == '=':
                    search_params["namespace"] = arg2
                else:
                    raise NotImplementedError("Only '=' operator is supported for namespace column.")
            elif arg1 == 'repository':
                if op == '=':
                    search_params["repository"] = arg2
                else:
                    raise NotImplementedError("Only '=' operator is supported for repository column.")
            elif arg1 in self.get_columns():
                subset_where_conditions.append([op, arg1, arg2])

        filter_flag = ("namespace" in search_params) and ("repository" in search_params)

        if not filter_flag:
            raise NotImplementedError("namespace and repository column has to be present in where clause.")

        repo_tags_summary_df = pd.DataFrame(columns=self.get_columns())

        response = self.handler.docker_client.get_repo_tags(search_params["namespace"], search_params["repository"])

        self.check_res(res=response)

        content = response["content"]

        repo_tags_summary_df = pd.json_normalize(content["results"])

        select_statement_executor = SELECTQueryExecutor(
            repo_tags_summary_df,
            selected_columns,
            subset_where_conditions,
            order_by_conditions,
            result_limit
        )

        repo_tags_summary_df = select_statement_executor.execute_query()

        return repo_tags_summary_df

    def check_res(self, res):
        if res["code"] != 200:
            raise Exception("Error fetching results - " + res["error"])

    def get_columns(self) -> List[str]:
        """Gets all columns to be returned in pandas DataFrame responses

        Returns
        -------
        List[str]
            List of columns
        """

        return ["creator", 
                "id", 
                "images", 
                "last_updated", 
                "last_updater", 
                "last_updater_username", 
                "name", 
                "repository", 
                "full_size", 
                "v2", 
                "tag_status", 
                "tag_last_pulled", 
                "tag_last_pushed", 
                "media_type", 
                "content_type"
            ]


import inspect
from typing import List
from dataclasses import dataclass

import pandas as pd
import github

from mindsdb.integrations.utilities.sql_utils import (FilterCondition, FilterOperator, SortColumn)
from mindsdb.integrations.libs.api_handler import APIResource


@dataclass
class Type:
    name: str
    sub_type: str = None
    optional: bool = False


@dataclass
class GHMethod:
    name: str
    table_name: str
    params: dict
    output: Type


def parse_annotations(annotations):
    '''
    Parse string annotation, and extract type, input examples:
    - Milestone | Opt[str]
    - PaginatedList[Issue]
    '''
    type_name, sub_type = None, None
    if not isinstance(annotations, str):

        return Type(getattr(annotations, '__name__', None))
    for item in annotations.split('|'):
        item = item.strip()
        if item is None:
            continue
        if '[' in item:
            type_name = item[: item.find('[')]
            item2 = item[item.find('[') + 1: item.rfind(']')]
            if type_name == 'Opt':
                inner_type = parse_annotations(item2)
                inner_type.optional = Type
                return inner_type
            if type_name == 'dict':
                item2 = item2[item2.find(',') + 1:]
            sub_type = parse_annotations(item2).name
        else:
            type_name = item
        # get only first type
        break
    return Type(type_name, sub_type)


def get_properties(cls):
    # find properties of the class

    properties = {}
    for prop_name, prop in inspect.getmembers(cls):
        if prop_name.startswith('_'):
            continue
        if not isinstance(prop, property):
            continue
        sig2 = inspect.signature(prop.fget)

        properties[prop_name] = parse_annotations(sig2.return_annotation)
    return properties


def get_github_types():
    # get github types
    types = {}

    GithubObject = github.GithubObject.GithubObject
    for module_name, module in inspect.getmembers(github, inspect.ismodule):
        cls = getattr(module, module_name, None)
        if cls is None:
            continue
        if issubclass(cls, GithubObject):

            # remove inherited props
            parent_props = []
            for cls2 in cls.__bases__:
                parent_props += get_properties(cls2).keys()

            properties = {}
            for k, v in get_properties(cls).items():
                if k not in parent_props:
                    properties[k] = v

            types[module_name] = properties
    return types


def get_github_methods(cls):
    '''
    Analyse class in order to find methods which return list of objects.
    '''
    methods = []

    for method_name, method in inspect.getmembers(cls, inspect.isfunction):
        sig = inspect.signature(method)

        return_type = parse_annotations(sig.return_annotation)
        list_prefix = 'get_'
        if not (method_name.startswith(list_prefix) and return_type.name == 'PaginatedList'):
            continue

        table_name = method_name[len(list_prefix):]

        params = {}
        for param_name, param in sig.parameters.items():
            params[param_name] = parse_annotations(param.annotation)

        methods.append(GHMethod(
            name=method_name,
            table_name=table_name,
            params=params,
            output=return_type
        ))
    return methods


class GHTable(APIResource):
    def __init__(self, *args, method: GHMethod = None, github_types=None, **kwargs):
        self.method = method
        self.github_types = github_types

        self.output_columns = {}
        if method.output.sub_type in self.github_types:
            self.output_columns = self.github_types[method.output.sub_type]

        # check params:
        self.params, self.list_params = [], []
        for name, param_type in method.params.items():
            self.params.append(name)
            if param_type.name == 'list':
                self.list_params.append(name)

        self._allow_sort = 'sort' in method.params

        super().__init__(*args, **kwargs)

    def repr_value(self, value, type_name):
        if value is None or type_name in ('bool', 'int', 'float'):
            return value
        if type_name in self.github_types:
            properties = self.github_types[type_name]
            if 'login' in properties:
                value = getattr(value, 'login')
            elif 'url' in properties:
                value = getattr(value, 'url')
        return str(value)

    def get_columns(self) -> List[str]:
        return list(self.output_columns.keys())

    def list(
        self,
        conditions: List[FilterCondition] = None,
        limit: int = None,
        sort: List[SortColumn] = None,
        targets: List[str] = None,
        **kwargs
    ) -> pd.DataFrame:

        if limit is None:
            limit = 20

        method_kwargs = {}
        if sort is not None and self._allow_sort:
            for col in sort:
                method_kwargs['sort'] = col.column
                method_kwargs['direction'] = 'asc' if col.ascending else 'desc'
                sort.applied = True
                # supported only 1 column
                break

        if conditions:
            for condition in conditions:
                if condition.column not in self.params:
                    continue

                if condition.column in self.list_params:
                    if condition.op == FilterOperator.IN:
                        method_kwargs[condition.column] = condition.value
                    elif condition.op == FilterOperator.EQUAL:
                        method_kwargs[condition.column] = [condition]
                    condition.applied = True
                else:
                    method_kwargs[condition.column] = condition.value
                    condition.applied = True

        connection = self.handler.connect()
        method = getattr(connection.get_repo(self.handler.repository), self.method.name)

        data = []
        count = 0
        for record in method(**method_kwargs):
            item = {}
            for name, output_type in self.output_columns.items():

                # workaround to prevent making addition request per property.
                if name in targets:
                    # request only if is required
                    value = getattr(record, name)
                else:
                    value = getattr(record, '_' + name).value
                if value is not None:
                    if output_type.name == 'list':
                        value = ",".join([
                                str(self.repr_value(i, output_type.sub_type))
                                for i in value
                        ])
                    else:
                        value = self.repr_value(value, output_type.name)
                item[name] = value

            data.append(item)

            count += 1
            if limit <= count:
                break

        return pd.DataFrame(data, columns=self.get_columns())



from dataclasses import dataclass
from typing import List, Optional

from mindsdb.api.executor.sql_query.result_set import ResultSet


@dataclass(kw_only=True, slots=True)
class ExecuteAnswer:
    data: Optional[ResultSet] = None
    state_track: Optional[List[List]] = None
    error_code: Optional[int] = None
    error_message: Optional[str] = None
    affected_rows: Optional[int] = None



import base64
from io import BytesIO
import os
from typing import Union
from urllib.parse import urlparse

import fitz  # PyMuPDF
from markitdown import MarkItDown
import mimetypes
from openai import OpenAI
import requests


class ToMarkdown:
    """
    Extracts the content of documents of various formats in markdown format.
    """
    def __init__(self, use_llm: bool, llm_client: OpenAI = None, llm_model: str = None):
        """
        Initializes the ToMarkdown class.
        """
        # If use_llm is True, llm_client and llm_model must be provided.
        if use_llm and (llm_client is None or llm_model is None):
            raise ValueError('LLM client and model must be provided when use_llm is True.')

        # If use_llm is False, set llm_client and llm_model to None even if they are provided.
        if not use_llm:
            llm_client = None
            llm_model = None

        # Only OpenAI is supported for now.
        # TODO: Add support for other LLMs.
        if llm_client is not None and not isinstance(llm_client, OpenAI):
            raise ValueError('Only OpenAI models are supported at the moment.')

        self.use_llm = use_llm
        self.llm_client = llm_client
        self.llm_model = llm_model

    def call(self, file_path_or_url: str) -> str:
        """
        Converts a file to markdown.
        """
        file_extension = self._get_file_extension(file_path_or_url)
        file = self._get_file_content(file_path_or_url)

        if file_extension == '.pdf':
            return self._pdf_to_markdown(file)
        elif file_extension in ['.jpg', '.jpeg', '.png', '.gif']:
            return self._image_to_markdown(file)
        else:
            return self._other_to_markdown(file)

    def _get_file_content(self, file_path_or_url: str) -> str:
        """
        Retrieves the content of a file.
        """
        parsed_url = urlparse(file_path_or_url)
        if parsed_url.scheme in ('http', 'https'):
            response = requests.get(file_path_or_url)
            if response.status_code == 200:
                return response
            else:
                raise RuntimeError(f'Unable to retrieve file from URL: {file_path_or_url}')
        else:
            with open(file_path_or_url, 'rb') as file:
                return BytesIO(file.read())

    def _get_file_extension(self, file_path_or_url: str) -> str:
        """
        Retrieves the file extension from a file path or URL.
        """
        parsed_url = urlparse(file_path_or_url)
        if parsed_url.scheme in ('http', 'https'):
            try:
                # Make a HEAD request to get headers without downloading the file.
                response = requests.head(file_path_or_url, allow_redirects=True)
                content_type = response.headers.get('Content-Type', '')
                if content_type:
                    ext = mimetypes.guess_extension(content_type.split(';')[0].strip())
                    if ext:
                        return ext

                # Fallback to extracting extension from the URL path
                ext = os.path.splitext(parsed_url.path)[1]
                if ext:
                    return ext
            except requests.RequestException:
                raise RuntimeError(f'Unable to retrieve file extension from URL: {file_path_or_url}')
        else:
            return os.path.splitext(file_path_or_url)[1]

    def _pdf_to_markdown(self, file_content: Union[requests.Response, bytes]) -> str:
        """
        Converts a PDF file to markdown.
        """
        if self.llm_client is None:
            return self._pdf_to_markdown_no_llm(file_content)
        else:
            return self._pdf_to_markdown_llm(file_content)

    def _pdf_to_markdown_llm(self, file_content: Union[requests.Response, BytesIO]) -> str:
        """
        Converts a PDF file to markdown using LLM.
        The LLM is used mainly for the purpose of generating descriptions of any images in the PDF.
        """
        if isinstance(file_content, requests.Response):
            file_content = BytesIO(file_content.content)

        document = fitz.open(stream=file_content, filetype="pdf")

        markdown_content = []
        for page_num in range(len(document)):
            page = document.load_page(page_num)

            # Get text blocks with coordinates.
            page_content = []
            blocks = page.get_text("blocks")
            for block in blocks:
                x0, y0, x1, y1, text, _, _ = block
                if text.strip():  # Skip empty or whitespace blocks.
                    page_content.append((y0, text.strip()))

            # Extract images from the page.
            image_list = page.get_images(full=True)
            for img_index, img in enumerate(image_list):
                xref = img[0]
                base_image = document.extract_image(xref)
                image_bytes = base_image["image"]

                # Use actual image y-coordinate if available.
                y0 = float(base_image.get("y", 0))
                image_description = self._generate_image_description(image_bytes)
                page_content.append((y0, f"![{image_description}](image_{page_num + 1}_{img_index + 1}.png)"))

            # Sort the content by y0 coordinate
            page_content.sort(key=lambda x: x[0])

            # Add sorted content to the markdown
            for _, text in page_content:
                markdown_content.append(text)
            markdown_content.append("\n")

        document.close()

        return "\n".join(markdown_content)

    def _generate_image_description(self, image_bytes: bytes) -> str:
        """
        Generates a description of the image using LLM.
        """
        image_base64 = base64.b64encode(image_bytes).decode("utf-8")

        response = self.llm_client.chat.completions.create(
            model=self.llm_model,
            messages=[
                {
                    "role": "user",
                    "content": [
                        {"type": "text", "text": "Describe this image"},
                        {"type": "image_url", "image_url": {"url": f"data:image/png;base64,{image_base64}"}},
                    ],
                }
            ],
        )
        description = response.choices[0].message.content
        return description

    def _pdf_to_markdown_no_llm(self, file_content: Union[requests.Response, BytesIO]) -> str:
        """
        Converts a PDF file to markdown without using LLM.
        """
        md = MarkItDown(enable_plugins=True)
        result = md.convert(file_content)
        return result.markdown

    def _image_to_markdown(self, file_content: Union[requests.Response, BytesIO]) -> str:
        """
        Converts images to markdown.
        """
        if not self.use_llm or self.llm_client is None:
            raise ValueError('LLM client must be enabled to convert images to markdown.')

        md = MarkItDown(llm_client=self.llm_client, llm_model=self.llm_model, enable_plugins=True)
        result = md.convert(file_content)
        return result.markdown

    def _other_to_markdown(self, file_content: Union[requests.Response, BytesIO]) -> str:
        """
        Converts other file formats to markdown.
        """
        md = MarkItDown(enable_plugins=True)
        result = md.convert(file_content)
        return result.markdown



from collections import OrderedDict

from mindsdb.integrations.libs.const import HANDLER_CONNECTION_ARG_TYPE as ARG_TYPE


connection_args = OrderedDict(
    user={
        'type': ARG_TYPE.STR,
        'description': 'User name',
        'required': True,
        'label': 'User'
    },
    password={
        'type': ARG_TYPE.PWD,
        'description': 'Password',
        'required': True,
        'label': 'Password',
        'secret': True
    },
    protocol_version={
        'type': ARG_TYPE.INT,
        'description': 'is not required, and default to 4.',
        'required': False,
        'label': 'Protocol version'
    },
    host={
        'type': ARG_TYPE.STR,
        'description': ' is the host name or IP address of the ScyllaDB.',
        'required': True,
        'label': 'Host'
    },
    port={
        'type': ARG_TYPE.INT,
        'description': 'Server port',
        'required': True,
        'label': 'Port'
    },
    keyspace={
        'type': ARG_TYPE.STR,
        'description': 'is the keyspace to connect to. It is a top level container for tables.',
        'required': True,
        'label': 'Keyspace'
    },
    secure_connect_bundle={
        'type': ARG_TYPE.STR,
        'description': 'Path or URL to the secure connect bundle',
        'required': True,
        'label': 'Host'
    }
)



from mindsdb.api.executor.planner.steps import MultipleSteps

from mindsdb.api.executor.exceptions import NotSupportedYet

from .base import BaseStepCall


class MultipleStepsCall(BaseStepCall):

    bind = MultipleSteps

    def call(self, step):

        if step.reduce != 'union':
            raise NotSupportedYet(f"Only MultipleSteps with type = 'union' is supported. Got '{step.type}'")
        data = None
        for substep in step.steps:
            subdata = self.sql_query.execute_step(substep)
            if data is None:
                data = subdata
            else:
                data.add_from_result_set(subdata)

        return data



from mindsdb.api.executor.data_types.response_type import RESPONSE_TYPE
from mindsdb.integrations.handlers.google_analytics_handler.google_analytics_handler import GoogleAnalyticsHandler

import unittest


class GoogleAnalyticsTest(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.kwargs = {
            "connection_data": {
                "credentials_file": '/home/talaat/Downloads/credentials.json',
                "property_id": '<YOUR_PROPERTY_ID>'
            }
        }

        cls.handler = GoogleAnalyticsHandler('test_google_analytics_handler', **cls.kwargs)

    def test_0_check_connection(self):
        assert self.handler.check_connection()

    def test_1_get_tables(self):
        tables = self.handler.get_tables()
        assert tables.type is RESPONSE_TYPE.TABLE

    def test_2_native_query_select(self):
        query = 'SELECT * FROM conversion_events'
        result = self.handler.native_query(query)
        assert result.type is RESPONSE_TYPE.TABLE

    def test_3_native_query_update(self):
        query = 'UPDATE conversion_events SET countingMethod = 1 WHERE name = "properties/371809744/conversionEvents/6637248600"'
        result = self.handler.native_query(query)
        assert result.type is RESPONSE_TYPE.OK

    def test_4_native_query_delete(self):
        query = 'DELETE FROM conversion_events WHERE name = "properties/371809744/conversionEvents/6622916665"'
        result = self.handler.native_query(query)
        assert result.type is RESPONSE_TYPE.OK

    def test_5_native_query_insert(self):
        query = "INSERT INTO conversion_events (event_name, countingMethod) VALUES ('event_4', 2)"
        result = self.handler.native_query(query)
        assert result.type is RESPONSE_TYPE.OK


if __name__ == '__main__':
    unittest.main()



import os
import pytest
import snowflake
import pandas as pd
import snowflake.connector
from mindsdb.integrations.handlers.snowflake_handler.snowflake_handler import SnowflakeHandler
from mindsdb.integrations.libs.response import (
    HandlerResponse as Response,
    RESPONSE_TYPE,
)

HANDLER_KWARGS = {
    "connection_data": {
        "account": os.environ.get("MDB_TEST_SNOWFLAKE_ACCOUNT"),
        "user": os.environ.get("MDB_TEST_SNOWFLAKE_USER"),
        "password": os.environ.get("MDB_TEST_SNOWFLAKE_PASSWORD"),
        "database": os.environ.get("MDB_TEST_SNOWFLAKE_DATABASE"),
        "schema": os.environ.get("MDB_TEST_SNOWFLAKE_SCHEMA", "PUBLIC"),
    }
}


@pytest.fixture(scope="class")
def snowflake_handler():
    """
    Create a SnowflakeHandler instance for testing.
    """

    seed_db()
    handler = SnowflakeHandler("test_snowflake_handler", **HANDLER_KWARGS)
    yield handler
    handler.disconnect()


def seed_db():
    """
    Seed the test DB by running the queries in the seed.sql file.
    """

    # Connect to the SNOWFLAKE database to run seed queries
    conn_info = HANDLER_KWARGS["connection_data"].copy()
    conn_info["database"] = "SNOWFLAKE"
    db = snowflake.connector.connect(**conn_info)

    cursor = db.cursor()
    with open("mindsdb/integrations/handlers/snowflake_handler/tests/seed.sql", "r") as f:
        for line in f.readlines():
            cursor.execute(line)
    cursor.close()
    db.close()


def check_valid_response(res):
    """
    Utility function to check if the response is valid.
    """

    if res.resp_type == RESPONSE_TYPE.TABLE:
        assert res.data_frame is not None, "expected to have some data, but got None"
    assert (
        res.error_code == 0
    ), f"expected to have zero error_code, but got {res.error_code}"
    assert (
        res.error_message is None
    ), f"expected to have None in error message, but got {res.error_message}"


def get_table_names(snowflake_handler):
    """
    Utility function to get the table names from the Snowflake account.
    """

    res = snowflake_handler.get_tables()
    tables = res.data_frame

    assert tables is not None, "expected to have some tables in the db, but got None"
    assert (
        "table_name" in tables
    ), f"expected to get 'table_name' column in the response:\n{tables}"

    return list(tables["table_name"])


@pytest.mark.usefixtures("snowflake_handler")
class TestSnowflakeHandlerConnect:
    def test_connect(self, snowflake_handler):
        """
        Tests the `connect` method to ensure it connects to the Snowflake account.
        """

        snowflake_handler.connect()
        assert snowflake_handler.is_connected, "the handler has failed to connect"

    def test_check_connection(self, snowflake_handler):
        """
        Tests the `check_connection` method to verify that it returns a StatusResponse object and accurately reflects the connection status.
        """

        res = snowflake_handler.check_connection()
        assert res.success, res.error_message


@pytest.mark.usefixtures("snowflake_handler")
class TestSnowflakeHandlerTables:
    table_for_creation = "TEST_MDB"

    def test_get_tables(self, snowflake_handler):
        """
        Tests the `get_tables` method to confirm it correctly calls `native_query` with the appropriate SQL commands.
        """

        res = snowflake_handler.get_tables()
        assert res.type == RESPONSE_TYPE.TABLE, "expected a TABLE"
        assert len(res.data_frame) > 0, "expected > O columns"

        tables = res.data_frame

        assert (
            tables is not None
        ), "expected to have some tables in the db, but got None"
        assert (
            "table_name" in tables
        ), f"expected to get 'table_name' in the response but got: {tables}"
        assert (
            "TEST" in tables["table_name"].values
        ), "expected to have 'test' in the response."

    def test_get_columns(self, snowflake_handler):
        """
        Tests if the `get_columns` method correctly constructs the SQL query and if it calls `native_query` with the correct query.
        """

        res = snowflake_handler.get_columns("TEST")
        assert res.type == RESPONSE_TYPE.TABLE, "expected a TABLE"
        assert len(res.data_frame) > 0, "expected > O columns"

        views = res.data_frame

        expected_columns = {
            "Field": ["COL_ONE", "COL_FOUR", "COL_TWO", "COL_THREE"],
            "Type": ["NUMBER", "TEXT", "NUMBER", "FLOAT"],
        }
        expected_df = pd.DataFrame(expected_columns)

        # Sort both DataFrames by all columns before comparing
        views = views.sort_values(by=list(res.data_frame.columns)).reset_index(drop=True)
        expected_df = expected_df.sort_values(by=list(expected_df.columns)).reset_index(drop=True)

        assert views.equals(
            expected_df
        ), "response does not contain the expected columns"

    def test_create_table(self, snowflake_handler):
        """
        Tests a table creation query to ensure it creates a table in the Snowflake account.
        """

        query = f"""
            CREATE TABLE IF NOT EXISTS {self.table_for_creation} (
                test_col INT
            );
        """
        res = snowflake_handler.native_query(query)
        check_valid_response(res)

        tables = get_table_names(snowflake_handler)

        assert (
            self.table_for_creation in tables
        ), f"expected to have {self.table_for_creation} in database, but got: {tables}"

    def test_drop_table(self, snowflake_handler):
        """
        Tests a table drop query to ensure it drops a table in the Snowflake account.
        """

        query = f"DROP TABLE IF EXISTS {self.table_for_creation}"
        res = snowflake_handler.native_query(query)
        check_valid_response(res)

        tables = get_table_names(snowflake_handler)

        assert self.table_for_creation not in tables


@pytest.mark.usefixtures("snowflake_handler")
class TestSnowflakeHandlerQuery:
    def test_select_native_query(self, snowflake_handler):
        """
        Tests the `native_query` method to ensure it executes a SQL query using a mock cursor and returns a Response object.
        """

        query = "SELECT * FROM test"
        res = snowflake_handler.native_query(query)

        assert type(res) is Response
        assert res.resp_type == RESPONSE_TYPE.TABLE

        expected_data = {
            "COL_ONE": [1, 2, 3],
            "COL_TWO": [-1, -2, -3],
            "COL_THREE": [0.1, 0.2, 0.3],
            "COL_FOUR": ["A", "B", "C"],
        }
        expected_df = pd.DataFrame(expected_data)

        assert res.data_frame.equals(
            expected_df
        ), "response does not contain the expected data"

    def test_select_query(self, snowflake_handler):
        """
        Tests the `query` method to ensure it executes a SQL query and returns a Response object.
        """

        limit = 3
        query = "SELECT * FROM test"
        res = snowflake_handler.query(query)
        check_valid_response(res)

        got_rows = res.data_frame.shape[0]
        want_rows = limit

        assert (
            got_rows == want_rows
        ), f"expected to have {want_rows} rows in response but got: {got_rows}"


if __name__ == "__main__":
    pytest.main([__file__])



import unittest
from mindsdb.integrations.handlers.dropbox_handler.dropbox_handler import DropboxHandler


class DropboxHandlerTest(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.kwargs = {
            "access_token": "ai.L-wqp3eP6r4cSWVklkKAdTNZ3VAuQjWuZMvIs1BzKvZNVW07rKbVNi5HbxvLc9q9D6qSfsf5VTsqYsNPGUkqSJBlpkr88gNboUNuhITmJG9mVw-Olniu4MO3BWVbEIphVxXxxxCd677Y",
        }
        cls.handler = DropboxHandler("test_dropbox_handler", cls.kwargs)

    def test_0_check_connection(self):
        assert self.handler.check_connection()


if __name__ == "__main__":
    unittest.main()



from collections import OrderedDict

from mindsdb.integrations.libs.const import HANDLER_CONNECTION_ARG_TYPE as ARG_TYPE


connection_args = OrderedDict(
    api_key={
        "type": ARG_TYPE.PWD,
        "description": "OilPriceAPI key to use for authentication.",
        "required": True,
        "label": "Api key",
        "secret": True
    }
)

connection_args_example = OrderedDict(
    api_key=""
)



from collections.abc import Callable

import redis
from pandas import DataFrame

from mindsdb.utilities.ml_task_queue.utils import RedisKey, from_bytes
from mindsdb.utilities.ml_task_queue.const import ML_TASK_STATUS


class Task:
    """ Abstraction for ML task. Should have interface similat to concurrent.futures.Future

        Attributes:
            db (Redis): database object
            redis_key (RedisKey): redis keys associated with task
            dataframe (DataFrame): task result
            exception (Exception): task exeuton  runtime exception
            _timeout (int): max time without status updating
    """

    def __init__(self, connection: redis.Redis, redis_key: RedisKey) -> None:
        self.db = connection
        self.redis_key = redis_key
        self.dataframe = None
        self.exception = None
        self._timeout = 30

    def subscribe(self) -> ML_TASK_STATUS:
        """ return tasks status untill it is not done or failed
        """
        pubsub = self.db.pubsub()
        cache = self.db.cache()
        pubsub.subscribe(self.redis_key.status)
        while (msg := pubsub.get_message(timeout=self._timeout)):
            if msg['type'] not in pubsub.PUBLISH_MESSAGE_TYPES:
                continue
            ml_task_status = ML_TASK_STATUS(msg['data'])
            if ml_task_status == ML_TASK_STATUS.COMPLETE:
                dataframe_bytes = cache.get(self.redis_key.dataframe)
                if dataframe_bytes is not None:
                    self.dataframe = from_bytes(dataframe_bytes)
                cache.delete(self.redis_key.dataframe)
            elif ml_task_status == ML_TASK_STATUS.ERROR:
                exception_bytes = cache.get(self.redis_key.exception)
                if exception_bytes is not None:
                    self.exception = from_bytes(exception_bytes)
            yield ml_task_status
        else:
            # there is no mesasges, timeout
            ml_task_status = ML_TASK_STATUS.TIMEOUT
            yield ml_task_status

    def wait(self, status: ML_TASK_STATUS = ML_TASK_STATUS.COMPLETE) -> None:
        """ block threasd untill task is not done or failed
        """
        for status in self.subscribe():
            if status in (ML_TASK_STATUS.WAITING, ML_TASK_STATUS.PROCESSING):
                continue
            if status == ML_TASK_STATUS.ERROR:
                if self.exception is not None:
                    raise self.exception
                else:
                    raise Exception('Unknown error during ML task execution')
            if status == ML_TASK_STATUS.TIMEOUT:
                raise Exception(f"Can't get answer in {self._timeout} seconds")
            if status == ML_TASK_STATUS.COMPLETE:
                return
            raise KeyError('Unknown task status')

    def result(self) -> DataFrame:
        """ wait task is done and return result

            Returns:
                DataFrame: task result
        """
        self.wait()
        return self.dataframe

    def add_done_callback(self, fn: Callable) -> None:
        """ need for compatability with concurrent.futures.Future interface
        """
        pass



import pandas as pd
from mindsdb.integrations.libs.api_handler import  APITable
from mindsdb.integrations.utilities.sql_utils import extract_comparison_conditions
from mindsdb_sql_parser import ast


class BalanceTable(APITable):
    '''A class representing the balance table.

    This class inherits from APITable and provides functionality to select data
    from the balance endpoint of the Plaid API and return it as a pandas DataFrame.

    Methods:
        select(ast.Select): Select data from the balance table and return it as a pandas DataFrame.
        get_columns(): Get the list of column names for the balance table.

    '''
    
    def select(self, query: ast.Select) :
        '''Select data from the balance table and return it as a pandas DataFrame.

        Args:
            query (ast.Select): The SQL query to be executed.

        Returns:
            pandas.DataFrame: A pandas DataFrame containing the selected data.
        '''

        conditions = extract_comparison_conditions(query.where)
        params = {}
        for i in conditions:
            if i[1] == 'last_updated_datetime':
                if i[0] == '=':
                    params[i[1]] = i[2]
                else:
                    raise Exception("Only equals to '=' is Supported with 'last_updated_datetime'")
                
        result = self.handler.call_plaid_api(method_name='get_balance', params=params)

        self.filter_columns(query=query, result=result)
        return result
    
    def get_columns(self):
        '''Get the list of column names for the balance table.
          
        Returns:
            list: A list of column names for the balance table.

        '''
        return [
            'account_id', 
            'account_name',
            'account_mask',
            'account_type',
            'account_subtype',
            'account_official_name',
            'balance_iso_currency_code',
            'balance_unofficial_currency_code',
            'balance_available',
            'balance_current',
            'balance_limit' 
        ]

    def filter_columns(self, result: pd.DataFrame, query: ast.Select = None):
        
        columns = []
        if query is not None:
            for target in query.targets:
                if isinstance(target, ast.Star):
                    columns = self.get_columns()
                    break
                elif isinstance(target, ast.Identifier):
                    columns.append(target.parts[-1])
                else:
                    raise NotImplementedError
        else:
            columns = self.get_columns()

        columns = [name.lower() for name in columns]

        if len(result) == 0:
            result = pd.DataFrame([], columns=columns)
        else:
            for col in set(columns) & set(result.columns) ^ set(columns):
                result[col] = None
            result = result[columns]

        if query is not None and query.limit is not None:
            return result.head(query.limit.value)

        return result


class TransactionTable(BalanceTable):
    '''A class representing the transaction table.

    This class inherits from APITable and provides functionality to select data
    from the transactions endpoint of the Plaid API and return it as a pandas DataFrame.

    Methods:
        select(ast.Select): Select data from the transaction table and return it as a pandas DataFrame.
        get_columns(): Get the list of column names for the transaction table.

    '''

    def select(self, query: ast.Select) :
        '''Select data from the transaction table and return it as a pandas DataFrame.

        Args:
            query (ast.Select): The SQL query to be executed.

        Returns:
            pandas.DataFrame: A pandas DataFrame containing the selected data.
        '''
        all_conditions = extract_comparison_conditions(query.where)
        condition = []
        params = {}
        for op, v, c in all_conditions:

            op = '==' if op == '=' else op  # converting '=' to '=='

            if (v == 'start_date' or v == 'end_date'):
                params[v] = c

            elif v in ['date', 'authorized_date'] or isinstance(c, str):
                condition.append(f"({v}{op}'{c}')")

            else:
                condition.append(f"({v}{op}{c})")

        merge_condition = ' and '.join(condition)

        result = self.handler.call_plaid_api(method_name='get_transactions', params=params)
        if merge_condition != '':
            result = result.query(merge_condition)

        result = self.filter_columns(query=query, result=result)
        return result

    def get_columns(self):
        '''Get the list of column names for the transaction table.

        Returns:
            list: A list of column names for the transaction table.
        '''
        return [
            'account_id',
            'transaction_id',
            'amount',
            'iso_currency_code',
            'check_number',
            'date',
            'authorized_date',
            'merchant_name',
            'payment_channel',
            'pending',
        ]

    


from typing import Optional

import pandas as pd
import duckdb

from mindsdb_sql_parser import parse_sql
from mindsdb.integrations.libs.base import DatabaseHandler

from mindsdb_sql_parser.ast.base import ASTNode

from mindsdb.utilities import log
from mindsdb.integrations.libs.response import (
    HandlerStatusResponse as StatusResponse,
    HandlerResponse as Response,
    RESPONSE_TYPE
)


logger = log.getLogger(__name__)


class SheetsHandler(DatabaseHandler):
    """
    This handler handles connection and execution of the Airtable statements.
    """

    name = 'sheets'

    def __init__(self, name: str, connection_data: Optional[dict], **kwargs):
        """
        Initialize the handler.
        Args:
            name (str): name of particular handler instance
            connection_data (dict): parameters for connecting to the database
            **kwargs: arbitrary keyword arguments.
        """
        super().__init__(name)
        self.parser = parse_sql
        self.dialect = 'sheets'
        self.connection_data = connection_data
        self.kwargs = kwargs

        self.connection = None
        self.is_connected = False

    def __del__(self):
        if self.is_connected is True:
            self.disconnect()

    def connect(self) -> StatusResponse:
        """
        Set up the connection required by the handler.
        Returns:
            HandlerStatusResponse
        """
        url = f"https://docs.google.com/spreadsheets/d/{self.connection_data['spreadsheet_id']}/gviz/tq?tqx=out:csv&sheet={self.connection_data['sheet_name']}"
        self.sheet = pd.read_csv(url, on_bad_lines='skip')
        self.connection = duckdb.connect()
        self.connection.register(self.connection_data['sheet_name'], self.sheet)
        self.is_connected = True

        return self.connection

    def disconnect(self):
        """
        Close any existing connections.
        """
        if self.is_connected is False:
            return

        self.connection.close()
        self.is_connected = False
        return self.is_connected

    def check_connection(self) -> StatusResponse:
        """
        Check connection to the handler.
        Returns:
            HandlerStatusResponse
        """
        response = StatusResponse(False)
        need_to_close = self.is_connected is False

        try:
            self.connect()
            response.success = True
        except Exception as e:
            logger.error(f'Error connecting to the Google Sheet with ID {self.connection_data["spreadsheet_id"]}, {e}!')
            response.error_message = str(e)
        finally:
            if response.success is True and need_to_close:
                self.disconnect()
            if response.success is False and self.is_connected is True:
                self.is_connected = False

        return response

    def native_query(self, query: str) -> StatusResponse:
        """
        Receive raw query and act upon it somehow.
        Args:
            query (str): query in native format
        Returns:
            HandlerResponse
        """

        need_to_close = self.is_connected is False
        connection = self.connect()
        try:
            result = connection.execute(query).fetchdf()
            if not result.empty:
                response = Response(
                    RESPONSE_TYPE.TABLE,
                    result
                )
            else:
                response = Response(RESPONSE_TYPE.OK)
                connection.commit()
        except Exception as e:
            logger.error(f'Error running query: {query} on the Google Sheet with ID {self.connection_data["spreadsheet_id"]}!')
            response = Response(
                RESPONSE_TYPE.ERROR,
                error_message=str(e)
            )

        if need_to_close is True:
            self.disconnect()

        return response

    def query(self, query: ASTNode) -> StatusResponse:
        """
        Receive query as AST (abstract syntax tree) and act upon it somehow.
        Args:
            query (ASTNode): sql query represented as AST. May be any kind
                of query: SELECT, INTSERT, DELETE, etc
        Returns:
            HandlerResponse
        """
        return self.native_query(query.to_string())

    def get_tables(self) -> StatusResponse:
        """
        Return list of entities that will be accessible as tables.
        Returns:
            HandlerResponse
        """
        response = Response(
            RESPONSE_TYPE.TABLE,
            data_frame=pd.DataFrame(
                [self.connection_data['sheet_name']],
                columns=['table_name']
            )
        )

        return response

    def get_columns(self) -> StatusResponse:
        """
        Returns a list of entity columns.
        Args:
            table_name (str): name of one of tables returned by self.get_tables()
        Returns:
            HandlerResponse
        """
        response = Response(
            RESPONSE_TYPE.TABLE,
            data_frame=pd.DataFrame(
                {
                    'column_name': list(self.sheet.columns),
                    'data_type': self.sheet.dtypes
                }
            )
        )

        return response



from os import path
from typing import Collection, Dict

import numpy as np
import pandas as pd
import requests

SERVICE_URL = r"https://pypistats.org"
API_BASE_URL = path.join(SERVICE_URL, "api/packages/")


class PyPI:
    def __init__(self, name: str, limit: int = None) -> None:
        """initializer method

        Args:
            name(str): package name
        """
        self.name: str = name
        self.limit = limit
        self.endpoint: str = path.join(API_BASE_URL, name)

    def recent(self, period: str = None) -> pd.DataFrame:
        """recent endpoint

        Args:
            period (str, optional): the desired `day` or `week` or `month` period. Defaults to None.

        Returns:
            pd.DataFrame: pandas dataframe
        """
        endpoint: str = path.join(self.endpoint, "recent")
        params: Dict = {}

        if period:
            params["period"] = period

        payload = requests.get(endpoint, params=params).json()["data"]

        df = self.__to_dataframe(payload, [0])

        return df

    def overall(self, mirrors: bool = None) -> pd.DataFrame:
        """overall endpoint

        Args:
            mirrors (bool, optional): filter by mirrors. Defaults to None.

        Returns:
            pd.DataFrame: pandas dataframe
        """
        endpoint: str = path.join(self.endpoint, "overall")
        params: Dict = {}

        if mirrors is not None:
            params["mirrors"] = str(mirrors).lower()

        payload = requests.get(endpoint, params=params).json()["data"]
        df = self.__to_dataframe(payload, limit=self.limit)

        return df

    def python_major(self, version: str = None) -> pd.DataFrame:
        """python major endpoint

        Args:
            version (str, optional): filter by the major version number. Defaults to None.

        Returns:
            pd.DataFrame: pandas dataframe
        """
        endpoint: str = path.join(self.endpoint, "python_major")
        params: Dict = {}

        if version is not None:
            params["version"] = version

        payload = requests.get(endpoint, params=params).json()["data"]
        df = self.__to_dataframe(payload, limit=self.limit)

        return df

    def python_minor(self, version: str = None) -> pd.DataFrame:
        """python minor endpoint

        Args:
            version (str, optional): filter by the minor.patch version number. Defaults to None.

        Returns:
            pd.DataFrame: pandas dataframe
        """
        endpoint: str = path.join(self.endpoint, "python_minor")
        params: Dict = {}

        if version is not None:
            params["version"] = version

        payload = requests.get(endpoint, params=params).json()["data"]
        df = self.__to_dataframe(payload, limit=self.limit)

        return df

    def system(self, os: str = None) -> pd.DataFrame:
        """system endpoint

        Args:
            os (str, optional): filter by the operating system. Defaults to None.

        Returns:
            pd.DataFrame: pandas dataframe
        """
        endpoint: str = path.join(self.endpoint, "system")
        params: Dict = {}

        if os is not None:
            params["os"] = os

        payload = requests.get(endpoint, params=params).json()["data"]
        df = self.__to_dataframe(payload, limit=self.limit)

        return df

    @staticmethod
    def __to_dataframe(
        json_data: Dict,
        index: Collection = None,
        limit: int = 20,
    ) -> pd.DataFrame:
        """_summary_

        Args:
            json_data (Dict): data
            index (Collection, optional): desired index. Defaults to None.
            limit (int, optional): limit the output coming from dataframe. Defaults to 20.

        Returns:
            pd.DataFrame: _description_
        """
        df = pd.DataFrame(json_data, index=index)
        df.replace("null", np.nan, inplace=True)
        df = df.dropna()

        return df.tail(limit)

    @classmethod
    def is_connected(cls) -> Dict:
        try:
            _ = requests.get(SERVICE_URL, timeout=5).raise_for_status()
            return {"status": True}
        except requests.exceptions.RequestException as e:
            return {"status": False, "message": e}



from mindsdb.integrations.libs.api_handler import APIResource
from mindsdb.integrations.utilities.sql_utils import FilterCondition, FilterOperator
import pandas as pd
from typing import List


class ListFilesTable(APIResource):

    def list(self,
             targets: List[str] = None,
             conditions: List[FilterCondition] = None,
             limit: int = None,
             *args, **kwargs) -> pd.DataFrame:

        buckets = None
        for condition in conditions:
            if condition.column == 'bucket':
                if condition.op == FilterOperator.IN:
                    buckets = condition.value
                elif condition.op == FilterOperator.EQUAL:
                    buckets = [condition.value]
                condition.applied = True

        data = []
        for obj in self.handler.get_objects(limit=limit, buckets=buckets):
            path = obj['Key']
            path = path.replace('`', '')
            item = {
                'path': path,
                'bucket': obj['Bucket'],
                'name': path[path.rfind('/') + 1:],
                'extension': path[path.rfind('.') + 1:]
            }

            data.append(item)

        return pd.DataFrame(data=data, columns=self.get_columns())

    def get_columns(self) -> List[str]:
        return ["path", "name", "extension", "bucket", "content"]


class FileTable(APIResource):

    def list(self, targets: List[str] = None, table_name=None, *args, **kwargs) -> pd.DataFrame:
        return self.handler.read_as_table(table_name)

    def add(self, data, table_name=None):
        df = pd.DataFrame(data)
        return self.handler.add_data_to_table(table_name, df)



import pandas as pd
from typing import Optional, Dict

from mindsdb.utilities import log
from mindsdb.integrations.libs.base import BaseMLEngine
from mindsdb.integrations.utilities.handler_utils import get_api_key
from mindsdb.integrations.libs.api_handler_exceptions import MissingConnectionParams

from mindsdb.integrations.handlers.twelve_labs_handler.settings import TwelveLabsHandlerModel
from mindsdb.integrations.handlers.twelve_labs_handler.twelve_labs_api_client import TwelveLabsAPIClient


logger = log.getLogger(__name__)


class TwelveLabsHandler(BaseMLEngine):
    """
    Twelve Labs API handler implementation.
    """

    name = 'twelve_labs'

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.generative = True

    @staticmethod
    def create_validation(target: str, args: Dict = None, **kwargs: Dict) -> None:
        """
        Validates the create arguments. This method is called when creating a new model, prior to calling the create() method.

        Parameters
        ----------
        target : str
            Name of the target column.

        args : Dict
            Arguments from the USING clause.

        kwargs : Dict
            Additional arguments.

        Raises
        ------
        MissingConnectionParams
            If a USING clause is not provided.

        ValueError
            If the parameters in the USING clause are invalid.
        """

        # check for USING clause
        if 'using' not in args:
            raise MissingConnectionParams("Twelve Labs engine requires a USING clause! Refer to its documentation for more details.")
        else:
            # get USING args
            args = args['using']
            # pass args to TwelveLabsHandlerModel for validation
            TwelveLabsHandlerModel(**args)

    def create(self, target: str, df: Optional[pd.DataFrame] = None, args: Optional[Dict] = None) -> None:
        """
        Creates a model for for interacting with the Twelve Labs API. This method is called when creating a new model.
        The following steps are performed:
            1. Create an index if it doesn't exist already.
            2. Create video indexing tasks for all video files or video urls.
            3. Poll for video indexing tasks to complete.

        Parameters
        ----------
        target : str
            Name of the target column.

        df : pd.DataFrame, Optional
            DataFrame containing the data to be used in creating the model. This can include the columns containing video urls or video files.

        args : Dict, Optional
            Arguments from the USING clause.
        """

        # get USING args and add target
        args = args['using']
        args['target'] = target

        # get api client and api key
        twelve_labs_api_client, api_key = self._get_api_client(args)

        # update args with api key
        args['twelve_labs_api_key'] = api_key

        # get index if it exists
        index_id = twelve_labs_api_client.get_index_by_name(index_name=args['index_name'])

        # create index if it doesn't exist
        if not index_id:
            logger.info(f"Index {args['index_name']} does not exist. Creating index.")
            index_id = twelve_labs_api_client.create_index(
                index_name=args['index_name'],
                engine_id=args['engine_id'] if 'engine_id' in args else None,
                index_options=args['index_options'],
                addons=args['addons'] if 'addons' in args else []
            )

        else:
            logger.info(f"Index {args['index_name']} already exists. Using existing index.")

        # store index_id in args
        args['index_id'] = index_id

        # initialize video_urls and video_files
        video_urls, video_files = None, None

        # create video indexing tasks for all video files or video urls
        # video urls will be given precedence
        # check if video_urls_column has been set and use it to get the video urls
        if 'video_urls_column' in args:
            logger.info("video_urls_column has been set, therefore, it will be given precedence.")
            video_urls = df[args['video_urls_column']].tolist()

        # else, check if video_files_column has been set and use it to get the video files
        elif 'video_files_column' in args:
            logger.info("video_urls_column has not been set, therefore, video_files_column will be used.")
            video_files = df[args['video_files_column']].tolist()

        # else, check if video_urls or video_files have been set and use them
        else:
            logger.info("video_urls_column and video_files_column have not been set, therefore, video_urls and video_files will be used.")
            video_urls = args['video_urls'] if 'video_urls' in args else None
            video_files = args['video_files'] if 'video_files' in args else None

        # if video_urls and video_files are not set, then raise an exception
        if not video_urls and not video_files:
            logger.error("Neither video_urls_column, video_files_column, video_urls nor video_files have been set.")
            raise RuntimeError("Neither video_urls_column, video_files_column, video_urls nor video_files have been set. Please set one of them.")

        task_ids = twelve_labs_api_client.create_video_indexing_tasks(
            index_id=index_id,
            video_urls=video_urls,
            video_files=video_files,
        )

        # poll for video indexing tasks to complete
        twelve_labs_api_client.poll_for_video_indexing_tasks(task_ids=task_ids)

        # store args in model_storage
        self.model_storage.json_set('args', args)

    def predict(self, df: Optional[pd.DataFrame] = None, args: Optional[Dict] = None) -> None:
        """
        Predicts the target column for the given data. This method is called when making predictions.

        Parameters
        ----------
        df : pd.DataFrame, Optional
            DataFrame containing the data to be used in making predictions. This can include the column containing the queries to be run against the index.

        args : Dict, Optional
            Additional arguments.

        """

        # get args from model_storage
        args = self.model_storage.json_get('args')

        # get api client
        twelve_labs_api_client, _ = self._get_api_client(args)

        # check if task is search
        if args['task'] == 'search':
            # get search query
            # TODO: support multiple queries
            query = df[args['search_query_column']].tolist()[0]

            # search for query in index
            data = twelve_labs_api_client.search_index(
                index_id=args['index_id'],
                query=query,
                search_options=args['search_options']
            )

            # TODO: pick only the necessary columns?
            # TODO: structure nested columns?
            # metadata = ['score', 'start', 'end', 'video_id', 'confidence']
            # df_metadata = pd.json_normalize(data, record_path='metadata', meta=metadata, record_prefix='metadata_')
            # df_modules = pd.json_normalize(data, record_path='modules', meta=metadata, record_prefix='modules_')
            # df_predictions = pd.merge(df_metadata, df_modules, on=metadata)
            # return df_predictions
            return pd.json_normalize(data).add_prefix(args['target'] + '_')

        # check if task is summarize
        elif args['task'] == 'summarization':
            # sumarize videos
            video_ids = df['video_id'].tolist()
            data = twelve_labs_api_client.summarize_videos(
                video_ids=video_ids,
                summarization_type=args['summarization_type'],
                prompt=args['prompt']
            )

            if args['summarization_type'] in ('chapter', 'highlight'):
                return pd.json_normalize(data, record_path=f"{args['summarization_type']}s", meta=['id']).add_prefix(args['target'] + '_')
            else:
                return pd.json_normalize(data).add_prefix(args['target'] + '_')

    def describe(self, attribute: Optional[str] = None) -> pd.DataFrame:
        """
        Describes the model. This method is called when describing the model.

        Parameters
        ----------
        attribute : str, Optional
            The attribute to describe.

        Returns
        -------
        pd.DataFrame
            DataFrame containing the description of the model.
        """

        if attribute == "args":
            args = self.model_storage.json_get("args")
            return pd.DataFrame(args.items(), columns=["key", "value"])

        elif attribute == "indexed_videos":
            # get api client
            twelve_labs_api_client, _ = self._get_api_client()

            # get videos indexed in the index
            index_name = self.model_storage.json_get("args").get("index_name")
            indexed_videos = twelve_labs_api_client.list_videos_in_index(index_name=index_name)

            # structure nested columns
            indexed_video_data = []
            for video in indexed_videos:
                video_data = video.copy()
                video_data.pop("metadata")
                video_data.update(video["metadata"])

                # convert engine_ids to string
                video_data['engine_ids'] = ", ".join(video_data['engine_ids'])

                indexed_video_data.append(video_data)

            df_videos = pd.DataFrame(indexed_video_data)

            # rename _id to video_id
            df_videos.rename(columns={"_id": "video_id"}, inplace=True)

            # MindsDB GUI fails to display NaN values, so we replace them with 0
            df_videos.fillna(0, inplace=True)
            return df_videos

        else:
            tables = ["args", "indexed_videos"]
            return pd.DataFrame(tables, columns=["tables"])

    def _get_api_client(self, args: Dict = None) -> TwelveLabsAPIClient:
        """
        Returns a TwelveLabsAPIClient instance.

        Parameters
        ----------
        args : Dict
            Arguments from the USING clause.

        Returns
        -------
        TwelveLabsAPIClient
            TwelveLabsAPIClient instance.
        """

        if not args:
            args = self.model_storage.json_get('args')

        # get api key
        api_key = get_api_key(
            api_name=self.name,
            create_args=args,
            engine_storage=self.engine_storage,
        )

        base_url = args.get('base_url', None)

        # initialize TwelveLabsAPIClient
        return TwelveLabsAPIClient(api_key=api_key, base_url=base_url), api_key



from collections import OrderedDict

from mindsdb.integrations.libs.const import HANDLER_CONNECTION_ARG_TYPE as ARG_TYPE


connection_args = OrderedDict(
    url={
        "type": ARG_TYPE.STR,
        "description": "The URI-Like connection string to the CKAN server. If provided, it will override the other connection arguments.",
        "required": True,
        "label": "URL",
    },
    api_key={
        "type": ARG_TYPE.STR,
        "description": "The API key used to authenticate with the CKAN server. For CKAN 2.10+ API tokens are supported.",
        "required": False,
        "label": "API Key/Token",
    },
)


connection_args_example = OrderedDict(url="https://data.gov.au/data/", api_key="my_api_key")



from typing import Optional, Literal
from dataclasses import dataclass, astuple, fields

import pandas as pd
from mindsdb_sql_parser.ast.base import ASTNode

from mindsdb.utilities import log
from mindsdb.utilities.config import config
from mindsdb.integrations.utilities.sql_utils import extract_comparison_conditions
from mindsdb.api.mysql.mysql_proxy.libs.constants.mysql import MYSQL_DATA_TYPE
from mindsdb.api.executor.datahub.classes.tables_row import (
    TABLES_ROW_TYPE,
    TablesRow,
)

logger = log.getLogger(__name__)


def _get_scope(query):
    databases, tables = None, None
    try:
        conditions = extract_comparison_conditions(query.where)
    except NotImplementedError:
        return databases, tables
    for op, arg1, arg2 in conditions:
        if op == '=':
            scope = [arg2]
        elif op == 'in':
            if not isinstance(arg2, list):
                arg2 = [arg2]
            scope = arg2
        else:
            continue

        if arg1.lower() == 'table_schema':
            databases = scope
        elif arg1.lower() == 'table_name':
            tables = scope
    return databases, tables


class Table:

    deletable: bool = False
    visible: bool = False
    kind: str = 'table'


class SchemataTable(Table):
    name = 'SCHEMATA'
    columns = [
        "CATALOG_NAME",
        "SCHEMA_NAME",
        "DEFAULT_CHARACTER_SET_NAME",
        "DEFAULT_COLLATION_NAME",
        "SQL_PATH",
    ]

    @classmethod
    def get_data(cls, inf_schema=None, **kwargs):

        databases_meta = inf_schema.session.database_controller.get_list()
        data = [
            ["def", x["name"], "utf8mb4", "utf8mb4_0900_ai_ci", None]
            for x in databases_meta
        ]

        df = pd.DataFrame(data, columns=cls.columns)
        return df


class TablesTable(Table):
    name = 'TABLES'

    columns = [
        "TABLE_CATALOG",
        "TABLE_SCHEMA",
        "TABLE_NAME",
        "TABLE_TYPE",
        "ENGINE",
        "VERSION",
        "ROW_FORMAT",
        "TABLE_ROWS",
        "AVG_ROW_LENGTH",
        "DATA_LENGTH",
        "MAX_DATA_LENGTH",
        "INDEX_LENGTH",
        "DATA_FREE",
        "AUTO_INCREMENT",
        "CREATE_TIME",
        "UPDATE_TIME",
        "CHECK_TIME",
        "TABLE_COLLATION",
        "CHECKSUM",
        "CREATE_OPTIONS",
        "TABLE_COMMENT",
    ]

    @classmethod
    def get_data(cls, query: ASTNode = None, inf_schema=None, **kwargs):

        databases, _ = _get_scope(query)

        data = []
        for name in inf_schema.tables.keys():
            if databases is not None and name not in databases:
                continue
            row = TablesRow(TABLE_TYPE=TABLES_ROW_TYPE.SYSTEM_VIEW, TABLE_NAME=name)
            data.append(row.to_list())

        for ds_name, ds in inf_schema.persis_datanodes.items():
            if databases is not None and ds_name not in databases:
                continue

            if hasattr(ds, 'get_tables_rows'):
                ds_tables = ds.get_tables_rows()
            else:
                ds_tables = ds.get_tables()
            if len(ds_tables) == 0:
                continue
            elif isinstance(ds_tables[0], dict):
                ds_tables = [
                    TablesRow(
                        TABLE_TYPE=TABLES_ROW_TYPE.BASE_TABLE, TABLE_NAME=x["name"]
                    )
                    for x in ds_tables
                ]
            elif (
                isinstance(ds_tables, list)
                and len(ds_tables) > 0
                and isinstance(ds_tables[0], str)
            ):
                ds_tables = [
                    TablesRow(TABLE_TYPE=TABLES_ROW_TYPE.BASE_TABLE, TABLE_NAME=x)
                    for x in ds_tables
                ]
            for row in ds_tables:
                row.TABLE_SCHEMA = ds_name
                data.append(row.to_list())

        for ds_name in inf_schema.get_integrations_names():
            if databases is not None and ds_name not in databases:
                continue

            try:
                ds = inf_schema.get(ds_name)
                ds_tables = ds.get_tables()
                for row in ds_tables:
                    row.TABLE_SCHEMA = ds_name
                    data.append(row.to_list())
            except Exception:
                logger.error(f"Can't get tables from '{ds_name}'")

        for project_name in inf_schema.get_projects_names():
            if databases is not None and project_name not in databases:
                continue

            project_dn = inf_schema.get(project_name)
            project_tables = project_dn.get_tables()
            for row in project_tables:
                row.TABLE_SCHEMA = project_name
                data.append(row.to_list())

        df = pd.DataFrame(data, columns=cls.columns)
        return df


@dataclass
class ColumnsTableRow:
    """Represents a row in the COLUMNS table.
    Fields description: https://dev.mysql.com/doc/refman/8.4/en/information-schema-columns-table.html
    NOTE: attrs order matter, don't change it.
    """
    TABLE_CATALOG: Literal['def'] = 'def'
    TABLE_SCHEMA: Optional[str] = None
    TABLE_NAME: Optional[str] = None
    COLUMN_NAME: Optional[str] = None
    ORDINAL_POSITION: int = 0
    COLUMN_DEFAULT: Optional[str] = None
    IS_NULLABLE: Literal['YES', 'NO'] = 'YES'
    DATA_TYPE: str = MYSQL_DATA_TYPE.VARCHAR.value
    CHARACTER_MAXIMUM_LENGTH: Optional[int] = None
    CHARACTER_OCTET_LENGTH: Optional[int] = None
    NUMERIC_PRECISION: Optional[int] = None
    NUMERIC_SCALE: Optional[int] = None
    DATETIME_PRECISION: Optional[int] = None
    CHARACTER_SET_NAME: Optional[str] = None
    COLLATION_NAME: Optional[str] = None
    COLUMN_TYPE: Optional[str] = None
    COLUMN_KEY: Optional[str] = None
    EXTRA: Optional[str] = None
    PRIVILEGES: str = 'select'
    COLUMN_COMMENT: Optional[str] = None
    GENERATION_EXPRESSION: Optional[str] = None

    def __post_init__(self):
        # region check mandatory fields
        mandatory_fields = ['TABLE_SCHEMA', 'TABLE_NAME', 'COLUMN_NAME']
        if any(getattr(self, field_name) is None for field_name in mandatory_fields):
            raise ValueError('One of mandatory fields is missed when creating ColumnsTableRow')
        # endregion

        # region set default values depend on type
        defaults = {
            'COLUMN_TYPE': self.DATA_TYPE
        }
        if MYSQL_DATA_TYPE(self.DATA_TYPE) in (
            MYSQL_DATA_TYPE.TIMESTAMP,
            MYSQL_DATA_TYPE.DATETIME,
            MYSQL_DATA_TYPE.DATE
        ):
            defaults = {
                'DATETIME_PRECISION': 0,
                'COLUMN_TYPE': self.DATA_TYPE
            }
        elif MYSQL_DATA_TYPE(self.DATA_TYPE) in (
            MYSQL_DATA_TYPE.FLOAT,
            MYSQL_DATA_TYPE.DOUBLE,
            MYSQL_DATA_TYPE.DECIMAL
        ):
            defaults = {
                'NUMERIC_PRECISION': 12,
                'NUMERIC_SCALE': 0,
                'COLUMN_TYPE': self.DATA_TYPE
            }
        elif MYSQL_DATA_TYPE(self.DATA_TYPE) in (
            MYSQL_DATA_TYPE.TINYINT,
            MYSQL_DATA_TYPE.SMALLINT,
            MYSQL_DATA_TYPE.MEDIUMINT,
            MYSQL_DATA_TYPE.INT,
            MYSQL_DATA_TYPE.BIGINT
        ):
            defaults = {
                'NUMERIC_PRECISION': 20,
                'NUMERIC_SCALE': 0,
                'COLUMN_TYPE': self.DATA_TYPE
            }
        elif MYSQL_DATA_TYPE(self.DATA_TYPE) is MYSQL_DATA_TYPE.VARCHAR:
            defaults = {
                'CHARACTER_MAXIMUM_LENGTH': 1024,
                'CHARACTER_OCTET_LENGTH': 3072,
                'CHARACTER_SET_NAME': 'utf8',
                'COLLATION_NAME': 'utf8_bin',
                'COLUMN_TYPE': 'varchar(1024)'
            }
        else:
            # show as MYSQL_DATA_TYPE.TEXT:
            defaults = {
                'CHARACTER_MAXIMUM_LENGTH': 65535,      # from https://bugs.mysql.com/bug.php?id=90685
                'CHARACTER_OCTET_LENGTH': 65535,        #
                'CHARACTER_SET_NAME': 'utf8',
                'COLLATION_NAME': 'utf8_bin',
                'COLUMN_TYPE': 'text'
            }

        for key, value in defaults.items():
            setattr(self, key, value)

        self.DATA_TYPE = self.DATA_TYPE.lower()
        self.COLUMN_TYPE = self.COLUMN_TYPE.lower()
        # endregion


class ColumnsTable(Table):
    name = 'COLUMNS'
    columns = [field.name for field in fields(ColumnsTableRow)]

    @classmethod
    def get_data(cls, inf_schema=None, query: ASTNode = None, **kwargs):
        result = []

        databases, tables_names = _get_scope(query)

        if databases is None:
            databases = [
                'information_schema',
                config.get('default_project'),
                'files'
            ]

        for db_name in databases:
            tables = {}
            if db_name == 'information_schema':
                for table_name, table in inf_schema.tables.items():
                    tables[table_name] = [
                        {'name': name} for name in table.columns
                    ]
            else:
                dn = inf_schema.get(db_name)
                if dn is None:
                    continue

                if tables_names is None:
                    tables_names = [t.TABLE_NAME for t in dn.get_tables()]
                for table_name in tables_names:
                    tables[table_name] = dn.get_table_columns(table_name)

            for table_name, table_columns in tables.items():
                for i, column in enumerate(table_columns):
                    column_name = column['name']
                    column_type = column.get('type', 'text')

                    # region infer type
                    if isinstance(column_type, MYSQL_DATA_TYPE) is False:
                        if column_type in ('double precision', 'real', 'numeric', 'float'):
                            column_type = MYSQL_DATA_TYPE.FLOAT
                        elif column_type in ('integer', 'smallint', 'int', 'bigint'):
                            column_type = MYSQL_DATA_TYPE.BIGINT
                        elif column_type in (
                            'timestamp without time zone',
                            'timestamp with time zone',
                            'date', 'timestamp'
                        ):
                            column_type = MYSQL_DATA_TYPE.DATETIME
                        else:
                            column_type = MYSQL_DATA_TYPE.VARCHAR
                    # endregion

                    column_row = astuple(
                        ColumnsTableRow(
                            TABLE_SCHEMA=db_name,
                            TABLE_NAME=table_name,
                            COLUMN_NAME=column_name,
                            DATA_TYPE=column_type.value,
                            ORDINAL_POSITION=i
                        )
                    )

                    result.append(column_row)

        df = pd.DataFrame(result, columns=cls.columns)
        return df


class EventsTable(Table):
    name = "EVENTS"

    columns = [
        "EVENT_CATALOG",
        "EVENT_SCHEMA",
        "EVENT_NAME",
        "DEFINER",
        "TIME_ZONE",
        "EVENT_BODY",
        "EVENT_DEFINITION",
        "EVENT_TYPE",
        "EXECUTE_AT",
        "INTERVAL_VALUE",
        "INTERVAL_FIELD",
        "SQL_MODE",
        "STARTS",
        "ENDS",
        "STATUS",
        "ON_COMPLETION",
        "CREATED",
        "LAST_ALTERED",
        "LAST_EXECUTED",
        "EVENT_COMMENT",
        "ORIGINATOR",
        "CHARACTER_SET_CLIENT",
        "COLLATION_CONNECTION",
        "DATABASE_COLLATION",
    ]


class RoutinesTable(Table):
    name = "ROUTINE"
    columns = [
        "SPECIFIC_NAME",
        "ROUTINE_CATALOG",
        "ROUTINE_SCHEMA",
        "ROUTINE_NAME",
        "ROUTINE_TYPE",
        "DATA_TYPE",
        "CHARACTER_MAXIMUM_LENGTH",
        "CHARACTER_OCTET_LENGTH",
        "NUMERIC_PRECISION",
        "NUMERIC_SCALE",
        "DATETIME_PRECISION",
        "CHARACTER_SET_NAME",
        "COLLATION_NAME",
        "DTD_IDENTIFIER",
        "ROUTINE_BODY",
        "ROUTINE_DEFINITION",
        "EXTERNAL_NAME",
        "EXTERNAL_LANGUAGE",
        "PARAMETER_STYLE",
        "IS_DETERMINISTIC",
        "SQL_DATA_ACCESS",
        "SQL_PATH",
        "SECURITY_TYPE",
        "CREATED",
        "LAST_ALTERED",
        "SQL_MODE",
        "ROUTINE_COMMENT",
        "DEFINER",
        "CHARACTER_SET_CLIENT",
        "COLLATION_CONNECTION",
        "DATABASE_COLLATION",
    ]


class PluginsTable(Table):
    name = "PLUGINS"
    columns = [
        "PLUGIN_NAME",
        "PLUGIN_VERSION",
        "PLUGIN_STATUS",
        "PLUGIN_TYPE",
        "PLUGIN_TYPE_VERSION",
        "PLUGIN_LIBRARY",
        "PLUGIN_LIBRARY_VERSION",
        "PLUGIN_AUTHOR",
        "PLUGIN_DESCRIPTION",
        "PLUGIN_LICENSE",
        "LOAD_OPTION",
        "PLUGIN_MATURITY",
        "PLUGIN_AUTH_VERSION",
    ]


class EnginesTable(Table):
    name = "ENGINES"
    columns = ["ENGINE", "SUPPORT", "COMMENT", "TRANSACTIONS", "XA", "SAVEPOINTS"]

    @classmethod
    def get_data(cls, **kwargs):
        data = [
            [
                "InnoDB",
                "DEFAULT",
                "Supports transactions, row-level locking, and foreign keys",
                "YES",
                "YES",
                "YES",
            ]
        ]

        df = pd.DataFrame(data, columns=cls.columns)
        return df


class KeyColumnUsageTable(Table):
    name = "KEY_COLUMN_USAGE"
    columns = [
        "CONSTRAINT_CATALOG",
        "CONSTRAINT_SCHEMA",
        "CONSTRAINT_NAME",
        "TABLE_CATALOG",
        "TABLE_SCHEMA",
        "TABLE_NAME",
        "COLUMN_NAME",
        "ORDINAL_POSITION",
        "POSITION_IN_UNIQUE_CONSTRAINT",
        "REFERENCED_TABLE_SCHEMA",
        "REFERENCED_TABLE_NAME",
        "REFERENCED_COLUMN_NAME",
    ]


class StatisticsTable(Table):
    name = "STATISTICS"
    columns = [
        "TABLE_CATALOG",
        "TABLE_SCHEMA",
        "TABLE_NAME",
        "NON_UNIQUE",
        "INDEX_SCHEMA",
        "INDEX_NAME",
        "SEQ_IN_INDEX",
        "COLUMN_NAME",
        "COLLATION",
        "CARDINALITY",
        "SUB_PART",
        "PACKED",
        "NULLABLE",
        "INDEX_TYPE",
        "COMMENT",
        "INDEX_COMMENT",
        "IS_VISIBLE",
        "EXPRESSION",
    ]


class CharacterSetsTable(Table):
    name = "CHARACTER_SETS"
    columns = [
        "CHARACTER_SET_NAME",
        "DEFAULT_COLLATE_NAME",
        "DESCRIPTION",
        "MAXLEN",
    ]

    @classmethod
    def get_data(cls, **kwargs):
        data = [
            ["utf8", "UTF-8 Unicode", "utf8_general_ci", 3],
            ["latin1", "cp1252 West European", "latin1_swedish_ci", 1],
            ["utf8mb4", "UTF-8 Unicode", "utf8mb4_general_ci", 4],
        ]

        df = pd.DataFrame(data, columns=cls.columns)
        return df


class CollationsTable(Table):
    name = "COLLATIONS"

    columns = [
        "COLLATION_NAME",
        "CHARACTER_SET_NAME",
        "ID",
        "IS_DEFAULT",
        "IS_COMPILED",
        "SORTLEN",
        "PAD_ATTRIBUTE",
    ]

    @classmethod
    def get_data(cls, **kwargs):
        data = [
            ["utf8_general_ci", "utf8", 33, "Yes", "Yes", 1, "PAD SPACE"],
            ["latin1_swedish_ci", "latin1", 8, "Yes", "Yes", 1, "PAD SPACE"],
        ]

        df = pd.DataFrame(data, columns=cls.columns)
        return df



from typing import Optional
from collections import OrderedDict

from mindsdb.interfaces.database.projects import ProjectController
import mindsdb.utilities.profiler as profiler
from mindsdb.utilities.config import config
from mindsdb.utilities.exception import EntityNotExistsError
from mindsdb.interfaces.database.log import LogDBController


class DatabaseController:
    def __init__(self):
        from mindsdb.interfaces.database.integrations import integration_controller
        self.integration_controller = integration_controller
        self.project_controller = ProjectController()

        self.logs_db_controller = LogDBController()
        self.information_schema_controller = None

    def delete(self, name: str):
        databases = self.get_dict()
        name = name.lower()
        if name not in databases:
            raise EntityNotExistsError('Database does not exists', name)
        db_type = databases[name]['type']
        if db_type == 'project':
            project = self.get_project(name)
            project.delete()
            return
        elif db_type == 'data':
            self.integration_controller.delete(name)
            return
        else:
            raise Exception(f"Database with type '{db_type}' cannot be deleted")

    @profiler.profile()
    def get_list(self, filter_type: Optional[str] = None, with_secrets: Optional[bool] = True):
        projects = self.project_controller.get_list()
        integrations = self.integration_controller.get_all(show_secrets=with_secrets)
        result = [{
            'name': 'information_schema',
            'type': 'system',
            'id': None,
            'engine': None,
            'visible': True,
            'deletable': False
        }, {
            'name': 'log',
            'type': 'system',
            'id': None,
            'engine': None,
            'visible': True,
            'deletable': False
        }]
        for x in projects:
            result.append({
                'name': x.name,
                'type': 'project',
                'id': x.id,
                'engine': None,
                'visible': True,
                'deletable': x.name.lower() != config.get('default_project')
            })
        for key, value in integrations.items():
            db_type = value.get('type', 'data')
            if db_type != 'ml':
                result.append({
                    'name': key,
                    'type': value.get('type', 'data'),
                    'id': value.get('id'),
                    'engine': value.get('engine'),
                    'class_type': value.get('class_type'),
                    'connection_data': value.get('connection_data'),
                    'visible': True,
                    'deletable': value.get('permanent', False) is False
                })

        if filter_type is not None:
            result = [x for x in result if x['type'] == filter_type]

        return result

    def get_dict(self, filter_type: Optional[str] = None):
        return OrderedDict(
            (
                x['name'].lower(),
                {
                    'type': x['type'],
                    'engine': x['engine'],
                    'id': x['id']
                }
            )
            for x in self.get_list(filter_type=filter_type)
        )

    def get_integration(self, integration_id):
        # get integration by id

        # TODO get directly from db?
        for rec in self.get_list():
            if rec['id'] == integration_id and rec['type'] == 'data':
                return {
                    'name': rec['name'],
                    'type': rec['type'],
                    'engine': rec['engine'],
                    'id': rec['id']
                }

    def exists(self, db_name: str) -> bool:
        return db_name.lower() in self.get_dict()

    def get_project(self, name: str):
        return self.project_controller.get(name=name)

    def get_system_db(self, name: str):
        if name == 'log':
            return self.logs_db_controller
        elif name == 'information_schema':
            from mindsdb.api.executor.controllers.session_controller import SessionController
            session = SessionController()
            return session.datahub
        else:
            raise Exception(f"Database '{name}' does not exists")



from mindsdb_sql_parser import parse_sql

from mindsdb.integrations.handlers.zendesk_handler.zendesk_tables import (
    ZendeskUsersTable,
    ZendeskTicketsTable,
    ZendeskTriggersTable,
    ZendeskActivitiesTable
)
from mindsdb.integrations.libs.api_handler import APIHandler
from mindsdb.integrations.libs.response import (
    HandlerStatusResponse as StatusResponse,
)
from mindsdb.utilities import log
import zenpy

logger = log.getLogger(__name__)


class ZendeskHandler(APIHandler):
    """The Zendesk handler implementation"""

    def __init__(self, name: str, **kwargs):
        """Initialize the zendesk handler.

        Parameters
        ----------
        name : str
            name of a handler instance
        """
        super().__init__(name)

        connection_data = kwargs.get("connection_data", {})
        self.connection_data = connection_data
        self.kwargs = kwargs
        self.zen_client = None
        self.is_connected = False

        self._register_table("users", ZendeskUsersTable(self))
        self._register_table("tickets", ZendeskTicketsTable(self))
        self._register_table("triggers", ZendeskTriggersTable(self))
        self._register_table("activities", ZendeskActivitiesTable(self))

    def connect(self) -> StatusResponse:
        """Set up the connection required by the handler.

        Returns
        -------
        StatusResponse
            connection object
        """
        resp = StatusResponse(False)
        self.zen_client = zenpy.Zenpy(subdomain=self.connection_data["sub_domain"], email=self.connection_data["email"], token=self.connection_data["api_key"])
        try:
            self.zen_client.users()
            self.is_connected = True
            resp.success = True
        except Exception as ex:
            resp.success = False
            resp.error_message = str(ex)
            self.is_connected = False
        return resp

    def check_connection(self) -> StatusResponse:
        """Check connection to the handler.

        Returns
        -------
        StatusResponse
            Status confirmation
        """
        response = self.connect()
        self.is_connected = response.success
        return response

    def native_query(self, query: str) -> StatusResponse:
        """Receive and process a raw query.

        Parameters
        ----------
        query : str
            query in a native format

        Returns
        -------
        StatusResponse
            Request status
        """
        ast = parse_sql(query)
        return self.query(ast)



import json
import datetime
from typing import Dict, List

import numpy as np
from sqlalchemy import (
    JSON,
    Boolean,
    Column,
    DateTime,
    Index,
    Integer,
    LargeBinary,
    Numeric,
    String,
    UniqueConstraint,
    create_engine,
    text,
    types
)
from sqlalchemy.exc import OperationalError
from sqlalchemy.orm import (
    Mapped,
    mapped_column,
    declarative_base,
    relationship,
    scoped_session,
    sessionmaker,
)
from sqlalchemy.sql.schema import ForeignKey

from mindsdb.utilities.json_encoder import CustomJSONEncoder
from mindsdb.utilities.config import config


class Base:
    __allow_unmapped__ = True


Base = declarative_base(cls=Base)

session, engine = None, None


def init(connection_str: str = None):
    global Base, session, engine
    if connection_str is None:
        connection_str = config['storage_db']
    base_args = {
        "pool_size": 30,
        "max_overflow": 200,
        "json_serializer": CustomJSONEncoder().encode,
    }
    engine = create_engine(connection_str, echo=False, **base_args)
    session = scoped_session(sessionmaker(bind=engine, autoflush=True))
    Base.query = session.query_property()


def serializable_insert(record: Base, try_count: int = 100):
    """Do serializeble insert. If fail - repeat it {try_count} times.

    Args:
        record (Base): sqlalchey record to insert
        try_count (int): count of tryes to insert record
    """
    commited = False
    while not commited:
        session.connection(execution_options={"isolation_level": "SERIALIZABLE"})
        if engine.name == "postgresql":
            session.execute(text("LOCK TABLE PREDICTOR IN EXCLUSIVE MODE"))
        session.add(record)
        try:
            session.commit()
        except OperationalError:
            # catch 'SerializationFailure' (it should be in str(e), but it may depend on engine)
            session.rollback()
            try_count += -1
            if try_count == 0:
                raise
        else:
            commited = True


# Source: https://stackoverflow.com/questions/26646362/numpy-array-is-not-json-serializable
class NumpyEncoder(json.JSONEncoder):
    """Special json encoder for numpy types"""

    def default(self, obj):
        if isinstance(obj, np.integer):
            return int(obj)
        elif isinstance(obj, np.floating):
            return float(obj)
        elif isinstance(obj, np.ndarray):
            return obj.tolist()
        return json.JSONEncoder.default(self, obj)


class Array(types.TypeDecorator):
    """Float Type that replaces commas with  dots on input"""

    impl = types.String

    def process_bind_param(self, value, dialect):  # insert
        if isinstance(value, str):
            return value
        elif value is None:
            return value
        else:
            return ",|,|,".join(value)

    def process_result_value(self, value, dialect):  # select
        return value.split(",|,|,") if value is not None else None


class Json(types.TypeDecorator):
    """Float Type that replaces commas with  dots on input"""

    impl = types.String

    def process_bind_param(self, value, dialect):  # insert
        return json.dumps(value, cls=NumpyEncoder) if value is not None else None

    def process_result_value(self, value, dialect):  # select
        if isinstance(value, dict):
            return value
        return json.loads(value) if value is not None else None


class PREDICTOR_STATUS:
    __slots__ = ()
    COMPLETE = "complete"
    TRAINING = "training"
    FINETUNING = "finetuning"
    GENERATING = "generating"
    ERROR = "error"
    VALIDATION = "validation"
    DELETED = "deleted"  # TODO remove it?


PREDICTOR_STATUS = PREDICTOR_STATUS()


class Predictor(Base):
    __tablename__ = "predictor"

    id = Column(Integer, primary_key=True)
    updated_at = Column(
        DateTime, default=datetime.datetime.now, onupdate=datetime.datetime.now
    )
    created_at = Column(DateTime, default=datetime.datetime.now)
    deleted_at = Column(DateTime)
    name = Column(String)
    data = Column(
        Json
    )  # A JSON -- should be everything returned by `get_model_data`, I think
    to_predict = Column(Array)
    company_id = Column(Integer)
    mindsdb_version = Column(String)
    native_version = Column(String)
    integration_id = Column(ForeignKey("integration.id", name="fk_integration_id"))
    data_integration_ref = Column(Json)
    fetch_data_query = Column(String)
    learn_args = Column(Json)
    update_status = Column(String, default="up_to_date")
    status = Column(String)
    active = Column(Boolean, default=True)
    training_data_columns_count = Column(Integer)
    training_data_rows_count = Column(Integer)
    training_start_at = Column(DateTime)
    training_stop_at = Column(DateTime)
    label = Column(String, nullable=True)
    version = Column(Integer, default=1)
    code = Column(String, nullable=True)
    lightwood_version = Column(String, nullable=True)
    dtype_dict = Column(Json, nullable=True)
    project_id = Column(
        Integer, ForeignKey("project.id", name="fk_project_id"), nullable=False
    )
    training_phase_current = Column(Integer)
    training_phase_total = Column(Integer)
    training_phase_name = Column(String)
    training_metadata = Column(JSON, default={}, nullable=False)

    @staticmethod
    def get_name_and_version(full_name):
        name_no_version = full_name
        version = None
        parts = full_name.split(".")
        if len(parts) > 1 and parts[-1].isdigit():
            version = int(parts[-1])
            name_no_version = ".".join(parts[:-1])
        return name_no_version, version


Index(
    "predictor_index",
    Predictor.company_id,
    Predictor.name,
    Predictor.version,
    Predictor.active,
    Predictor.deleted_at,  # would be good to have here nullsfirst(Predictor.deleted_at)
    unique=True
)


class Project(Base):
    __tablename__ = "project"

    id = Column(Integer, primary_key=True)
    created_at = Column(DateTime, default=datetime.datetime.now)
    updated_at = Column(
        DateTime, default=datetime.datetime.now, onupdate=datetime.datetime.now
    )
    deleted_at = Column(DateTime)
    name = Column(String, nullable=False)
    company_id = Column(Integer, default=0)
    metadata_: dict = Column("metadata", JSON, nullable=True)
    __table_args__ = (
        UniqueConstraint("name", "company_id", name="unique_project_name_company_id"),
    )


class Integration(Base):
    __tablename__ = "integration"
    id = Column(Integer, primary_key=True)
    updated_at = Column(
        DateTime, default=datetime.datetime.now, onupdate=datetime.datetime.now
    )
    created_at = Column(DateTime, default=datetime.datetime.now)
    name = Column(String, nullable=False)
    engine = Column(String, nullable=False)
    data = Column(Json)
    company_id = Column(Integer)
    __table_args__ = (
        UniqueConstraint(
            "name", "company_id", name="unique_integration_name_company_id"
        ),
    )


class File(Base):
    __tablename__ = "file"
    id = Column(Integer, primary_key=True)
    name = Column(String, nullable=False)
    company_id = Column(Integer)
    source_file_path = Column(String, nullable=False)
    file_path = Column(String, nullable=False)
    row_count = Column(Integer, nullable=False)
    columns = Column(Json, nullable=False)
    created_at = Column(DateTime, default=datetime.datetime.now)
    metadata_: dict = Column("metadata", JSON, nullable=True)
    updated_at = Column(
        DateTime, default=datetime.datetime.now, onupdate=datetime.datetime.now
    )
    __table_args__ = (
        UniqueConstraint("name", "company_id", name="unique_file_name_company_id"),
    )


class View(Base):
    __tablename__ = "view"
    id = Column(Integer, primary_key=True)
    name = Column(String, nullable=False)
    company_id = Column(Integer)
    query = Column(String, nullable=False)
    project_id = Column(
        Integer, ForeignKey("project.id", name="fk_project_id"), nullable=False
    )
    __table_args__ = (
        UniqueConstraint("name", "company_id", name="unique_view_name_company_id"),
    )


class JsonStorage(Base):
    __tablename__ = "json_storage"
    id = Column(Integer, primary_key=True)
    resource_group = Column(String)
    resource_id = Column(Integer)
    name = Column(String)
    content = Column(JSON)
    encrypted_content = Column(LargeBinary, nullable=True)
    company_id = Column(Integer)

    def to_dict(self) -> Dict:
        return {
            "id": self.id,
            "resource_group": self.resource_group,
            "resource_id": self.resource_id,
            "name": self.name,
            "content": self.content,
            "encrypted_content": self.encrypted_content,
            "company_id": self.company_id,
        }


class Jobs(Base):
    __tablename__ = "jobs"
    id = Column(Integer, primary_key=True)
    company_id = Column(Integer)
    user_class = Column(Integer, nullable=True)
    active = Column(Boolean, default=True)

    name = Column(String, nullable=False)
    project_id = Column(Integer, nullable=False)
    query_str = Column(String, nullable=False)
    if_query_str = Column(String, nullable=True)
    start_at = Column(DateTime, default=datetime.datetime.now)
    end_at = Column(DateTime)
    next_run_at = Column(DateTime)
    schedule_str = Column(String)

    deleted_at = Column(DateTime)
    updated_at = Column(
        DateTime, default=datetime.datetime.now, onupdate=datetime.datetime.now
    )
    created_at = Column(DateTime, default=datetime.datetime.now)


class JobsHistory(Base):
    __tablename__ = "jobs_history"
    id = Column(Integer, primary_key=True)
    company_id = Column(Integer)

    job_id = Column(Integer)

    query_str = Column(String)
    start_at = Column(DateTime)
    end_at = Column(DateTime)

    error = Column(String)
    created_at = Column(DateTime, default=datetime.datetime.now)
    updated_at = Column(DateTime, default=datetime.datetime.now)

    __table_args__ = (
        UniqueConstraint("job_id", "start_at", name="uniq_job_history_job_id_start"),
    )


class ChatBots(Base):
    __tablename__ = "chat_bots"
    id = Column(Integer, primary_key=True)

    name = Column(String, nullable=False)
    project_id = Column(Integer, nullable=False)
    agent_id = Column(ForeignKey("agents.id", name="fk_agent_id"))

    # To be removed when existing chatbots are backfilled with newly created Agents.
    model_name = Column(String)
    database_id = Column(Integer)
    params = Column(JSON)

    updated_at = Column(
        DateTime, default=datetime.datetime.now, onupdate=datetime.datetime.now
    )
    created_at = Column(DateTime, default=datetime.datetime.now)
    webhook_token = Column(String)

    def as_dict(self) -> Dict:
        return {
            "id": self.id,
            "name": self.name,
            "project_id": self.project_id,
            "agent_id": self.agent_id,
            "model_name": self.model_name,
            "params": self.params,
            "webhook_token": self.webhook_token,
            "created_at": self.created_at,
            "database_id": self.database_id,
        }


class ChatBotsHistory(Base):
    __tablename__ = "chat_bots_history"
    id = Column(Integer, primary_key=True)
    chat_bot_id = Column(Integer, nullable=False)
    type = Column(String)  # TODO replace to enum
    text = Column(String)
    user = Column(String)
    destination = Column(String)
    sent_at = Column(DateTime, default=datetime.datetime.now)
    error = Column(String)


class Triggers(Base):
    __tablename__ = "triggers"
    id = Column(Integer, primary_key=True)

    name = Column(String, nullable=False)
    project_id = Column(Integer, nullable=False)

    database_id = Column(Integer, nullable=False)
    table_name = Column(String, nullable=False)
    query_str = Column(String, nullable=False)
    columns = Column(String)  # list of columns separated by delimiter

    updated_at = Column(
        DateTime, default=datetime.datetime.now, onupdate=datetime.datetime.now
    )
    created_at = Column(DateTime, default=datetime.datetime.now)


class Tasks(Base):
    __tablename__ = "tasks"
    id = Column(Integer, primary_key=True)
    company_id = Column(Integer)
    user_class = Column(Integer, nullable=True)

    # trigger, chatbot
    object_type = Column(String, nullable=False)
    object_id = Column(Integer, nullable=False)

    last_error = Column(String)
    active = Column(Boolean, default=True)
    reload = Column(Boolean, default=False)

    # for running in concurrent processes
    run_by = Column(String)
    alive_time = Column(DateTime(timezone=True))

    updated_at = Column(
        DateTime, default=datetime.datetime.now, onupdate=datetime.datetime.now
    )
    created_at = Column(DateTime, default=datetime.datetime.now)


class AgentSkillsAssociation(Base):
    __tablename__ = "agent_skills"

    agent_id: Mapped[int] = mapped_column(ForeignKey("agents.id"), primary_key=True)
    skill_id: Mapped[int] = mapped_column(ForeignKey("skills.id"), primary_key=True)
    parameters: Mapped[dict] = mapped_column(JSON, default={}, nullable=True)

    agent = relationship("Agents", back_populates="skills_relationships")
    skill = relationship("Skills", back_populates="agents_relationships")


class Skills(Base):
    __tablename__ = "skills"
    id = Column(Integer, primary_key=True)
    agents_relationships: Mapped[List["Agents"]] = relationship(AgentSkillsAssociation, back_populates="skill")
    name = Column(String, nullable=False)
    project_id = Column(Integer, nullable=False)
    type = Column(String, nullable=False)
    params = Column(JSON)

    created_at = Column(DateTime, default=datetime.datetime.now)
    updated_at = Column(
        DateTime, default=datetime.datetime.now, onupdate=datetime.datetime.now
    )
    deleted_at = Column(DateTime)

    def as_dict(self) -> Dict:
        return {
            "id": self.id,
            "name": self.name,
            "project_id": self.project_id,
            "agent_ids": [rel.agent.id for rel in self.agents_relationships],
            "type": self.type,
            "params": self.params,
            "created_at": self.created_at,
        }


class Agents(Base):
    __tablename__ = "agents"
    id = Column(Integer, primary_key=True)
    skills_relationships: Mapped[List["Skills"]] = relationship(AgentSkillsAssociation, back_populates="agent")
    company_id = Column(Integer, nullable=True)
    user_class = Column(Integer, nullable=True)

    name = Column(String, nullable=False)
    project_id = Column(Integer, nullable=False)

    model_name = Column(String, nullable=True)
    provider = Column(String, nullable=True)
    params = Column(JSON)

    updated_at = Column(
        DateTime, default=datetime.datetime.now, onupdate=datetime.datetime.now
    )
    created_at = Column(DateTime, default=datetime.datetime.now)
    deleted_at = Column(DateTime)

    def as_dict(self) -> Dict:
        return {
            "id": self.id,
            "name": self.name,
            "project_id": self.project_id,
            "model_name": self.model_name,
            "skills": [rel.skill.as_dict() for rel in self.skills_relationships],
            "skills_extra_parameters": {rel.skill.name: (rel.parameters or {}) for rel in self.skills_relationships},
            "provider": self.provider,
            "params": self.params,
            "updated_at": self.updated_at,
            "created_at": self.created_at,
        }


class KnowledgeBase(Base):
    __tablename__ = "knowledge_base"
    id = Column(Integer, primary_key=True)
    name = Column(String, nullable=False)
    project_id = Column(Integer, nullable=False)
    params = Column(JSON)

    vector_database_id = Column(
        ForeignKey("integration.id", name="fk_knowledge_base_vector_database_id"),
        doc="fk to the vector database integration",
    )
    vector_database = relationship(
        "Integration",
        foreign_keys=[vector_database_id],
        doc="vector database integration",
    )

    vector_database_table = Column(String, doc="table name in the vector database")

    embedding_model_id = Column(
        ForeignKey("predictor.id", name="fk_knowledge_base_embedding_model_id"),
        doc="fk to the embedding model",
    )

    embedding_model = relationship(
        "Predictor", foreign_keys=[embedding_model_id], doc="embedding model"
    )
    query_id = Column(Integer, nullable=True)

    created_at = Column(DateTime, default=datetime.datetime.now)
    updated_at = Column(
        DateTime, default=datetime.datetime.now, onupdate=datetime.datetime.now
    )

    __table_args__ = (
        UniqueConstraint(
            "name", "project_id", name="unique_knowledge_base_name_project_id"
        ),
    )

    def as_dict(self) -> Dict:
        return {
            "id": self.id,
            "name": self.name,
            "project_id": self.project_id,
            "embedding_model": None if self.embedding_model is None else self.embedding_model.name,
            "vector_database": None if self.vector_database is None else self.vector_database.name,
            "vector_database_table": self.vector_database_table,
            "updated_at": self.updated_at,
            "created_at": self.created_at,
            "params": self.params
        }


class QueryContext(Base):
    __tablename__ = "query_context"
    id: int = Column(Integer, primary_key=True)
    company_id: int = Column(Integer, nullable=True)

    query: str = Column(String, nullable=False)
    context_name: str = Column(String, nullable=False)
    values: dict = Column(JSON)

    updated_at: datetime.datetime = Column(
        DateTime, default=datetime.datetime.now, onupdate=datetime.datetime.now
    )
    created_at: datetime.datetime = Column(DateTime, default=datetime.datetime.now)


class Queries(Base):
    __tablename__ = "queries"
    id: int = Column(Integer, primary_key=True)
    company_id: int = Column(Integer, nullable=True)

    sql: str = Column(String, nullable=False)
    # step_data: JSON = Column(JSON, nullable=True)

    started_at: datetime.datetime = Column(DateTime, default=datetime.datetime.now)
    finished_at: datetime.datetime = Column(DateTime)

    parameters = Column(JSON, default={})
    context = Column(JSON, default={})
    processed_rows = Column(Integer, default=0)
    error: str = Column(String, nullable=True)

    updated_at: datetime.datetime = Column(
        DateTime, default=datetime.datetime.now, onupdate=datetime.datetime.now
    )
    created_at: datetime.datetime = Column(DateTime, default=datetime.datetime.now)


class LLMLog(Base):
    __tablename__ = "llm_log"
    id: int = Column(Integer, primary_key=True)
    company_id: int = Column(Integer, nullable=False)
    api_key: str = Column(String, nullable=True)
    model_id: int = Column(Integer, nullable=True)
    model_group: str = Column(String, nullable=True)
    input: str = Column(JSON, nullable=True)
    output: str = Column(JSON, nullable=True)
    start_time: datetime = Column(DateTime, nullable=False)
    end_time: datetime = Column(DateTime, nullable=True)
    cost: float = Column(Numeric(5, 2), nullable=True)
    prompt_tokens: int = Column(Integer, nullable=True)
    completion_tokens: int = Column(Integer, nullable=True)
    total_tokens: int = Column(Integer, nullable=True)
    success: bool = Column(Boolean, nullable=False, default=True)
    exception: str = Column(String, nullable=True)
    traceback: str = Column(String, nullable=True)
    stream: bool = Column(Boolean, default=False, comment="Is this completion done in 'streaming' mode")
    metadata_: dict = Column("metadata", JSON, nullable=True)


class LLMData(Base):
    '''
    Stores the question/answer pairs of an LLM call so examples can be used
    for self improvement with DSPy
    '''
    __tablename__ = "llm_data"
    id: int = Column(Integer, primary_key=True)
    input: str = Column(String, nullable=False)
    output: str = Column(String, nullable=False)
    model_id: int = Column(Integer, nullable=False)
    created_at: datetime = Column(DateTime, default=datetime.datetime.now)
    updated_at: datetime = Column(DateTime, onupdate=datetime.datetime.now)



import os
import shutil
import tempfile
from pathlib import Path
from zipfile import ZipFile

import requests
from packaging.version import Version

from mindsdb.utilities.config import Config
from mindsdb.utilities import log

logger = log.getLogger(__name__)


def download_gui(destignation, version):
    if isinstance(destignation, str):
        destignation = Path(destignation)
    dist_zip_path = str(destignation.joinpath("dist.zip"))
    bucket = "https://mindsdb-web-builds.s3.amazonaws.com/"

    resources = [{"url": bucket + "dist-V" + version + ".zip", "path": dist_zip_path}]

    def get_resources(resource):
        response = requests.get(resource["url"])
        if response.status_code != requests.status_codes.codes.ok:
            raise Exception(f"Error {response.status_code} GET {resource['url']}")
        open(resource["path"], "wb").write(response.content)

    try:
        for r in resources:
            get_resources(r)
    except Exception as e:
        logger.error(f"Error during downloading files from s3: {e}")
        return False

    static_folder = destignation
    static_folder.mkdir(mode=0o777, exist_ok=True, parents=True)
    ZipFile(dist_zip_path).extractall(static_folder)

    if static_folder.joinpath("dist").is_dir():
        shutil.move(
            str(destignation.joinpath("dist").joinpath("index.html")), static_folder
        )
        shutil.move(
            str(destignation.joinpath("dist").joinpath("assets")), static_folder
        )
        shutil.rmtree(destignation.joinpath("dist"))

    os.remove(dist_zip_path)

    version_txt_path = destignation.joinpath("version.txt")
    with open(version_txt_path, "wt") as f:
        f.write(version)

    return True

    """
    # to make downloading faster download each resource in a separate thread
    with concurrent.futures.ThreadPoolExecutor(max_workers=5) as executor:
        future_to_url = {executor.submit(get_resources, r): r for r in resources}
        for future in concurrent.futures.as_completed(future_to_url):
            res = future.result()
            if res is not None:
                raise res
    """


def update_static(gui_version: Version):
    """Update Scout files basing on compatible-config.json content.
    Files will be downloaded and updated if new version of GUI > current.
    Current GUI version stored in static/version.txt.
    """
    config = Config()
    static_path = Path(config["paths"]["static"])

    logger.info(
        f"New version of GUI available ({gui_version.base_version}). Downloading..."
    )

    temp_dir = tempfile.mkdtemp(prefix="mindsdb_gui_files_")
    success = download_gui(temp_dir, gui_version.base_version)
    if success is False:
        shutil.rmtree(temp_dir)
        return False

    temp_dir_for_rm = tempfile.mkdtemp(prefix="mindsdb_gui_files_")
    shutil.rmtree(temp_dir_for_rm)
    shutil.copytree(str(static_path), temp_dir_for_rm)
    shutil.rmtree(str(static_path))
    shutil.copytree(temp_dir, str(static_path))
    shutil.rmtree(temp_dir_for_rm)

    logger.info(f"GUI version updated to {gui_version.base_version}")
    return True



from typing import Optional, Dict

import pandas as pd
import openai
from llama_index.llms.openai import OpenAI
from llama_index.core import Document
from llama_index.readers.web import SimpleWebPageReader
from llama_index.core import PromptTemplate
from llama_index.core import StorageContext, load_index_from_storage
from llama_index.embeddings.openai import OpenAIEmbedding
from llama_index.core import VectorStoreIndex
from llama_index.core import Settings

from mindsdb.integrations.libs.base import BaseMLEngine
from mindsdb.utilities.config import Config
from mindsdb.utilities.security import validate_urls
from mindsdb.integrations.handlers.llama_index_handler.settings import llama_index_config, LlamaIndexModel
from mindsdb.integrations.libs.api_handler_exceptions import MissingConnectionParams
from mindsdb.integrations.utilities.handler_utils import get_api_key


class LlamaIndexHandler(BaseMLEngine):
    """Integration with the LlamaIndex data framework for LLM applications."""

    name = "llama_index"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.generative = True
        self.default_index_class = llama_index_config.DEFAULT_INDEX_CLASS
        self.supported_index_class = llama_index_config.SUPPORTED_INDEXES
        self.default_reader = llama_index_config.DEFAULT_READER
        self.supported_reader = llama_index_config.SUPPORTED_READERS
        self.config = Config()

    @staticmethod
    def create_validation(target, args=None, **kwargs):
        if 'using' not in args:
            raise MissingConnectionParams("LlamaIndex engine requires USING clause!")
        else:
            args = args['using']
            LlamaIndexModel(**args)

    def create(
        self,
        target: str,
        df: Optional[pd.DataFrame] = None,
        args: Optional[Dict] = None,
    ) -> None:
        # workaround to create llama model without input data
        if df is None or df.empty:
            df = pd.DataFrame([{"text": ""}])
        
        args_reader = args.get("using", {}).get("reader", self.default_reader)

        if args_reader == "DFReader":
            dstrs = df.apply(
                lambda x: ", ".join(
                    [f"{col}: {str(entry)}" for col, entry in zip(df.columns, x)]
                ),
                axis=1,
            )
            reader = list(map(lambda x: Document(text=x), dstrs.tolist()))
        elif args_reader == "SimpleWebPageReader":
            url = args["using"]["source_url_link"]
            allowed_urls = self.config.get('web_crawling_allowed_sites', [])
            if allowed_urls and not validate_urls(url, allowed_urls):
                raise ValueError(f"The provided URL is not allowed for web crawling. Please use any of {', '.join(allowed_urls)}.")
            reader = SimpleWebPageReader(html_to_text=True).load_data([url])
        else:
            raise Exception(
                f"Invalid operation mode. Please use one of {self.supported_reader}."
            )
        self.model_storage.json_set("args", args)
        index = self._setup_index(reader)
        path = self.model_storage.folder_get("context")
        index.storage_context.persist(persist_dir=path)
        self.model_storage.folder_sync("context")

    def update(self, args) -> None:
        args_cur = self.model_storage.json_get("args")
        args_cur["using"].update(args["using"])

        # check new set of arguments
        self.create_validation(None, args_cur)

        self.model_storage.json_set("args", args_cur)

    def predict(
        self, df: Optional[pd.DataFrame] = None, args: Optional[Dict] = None
    ) -> pd.DataFrame:
        pred_args = args["predict_params"] if args else {}

        args = self.model_storage.json_get("args")
        engine_kwargs = {}

        if args["using"].get("mode") == "conversational":
            user_column = args["using"]["user_column"]
            assistant_column = args["using"]["assistant_column"]

            messages = []
            for row in df[:-1].to_dict("records"):
                messages.append(f"user: {row[user_column]}")
                messages.append(f"assistant: {row[assistant_column]}")

            conversation = "\n".join(messages)

            questions = [df.iloc[-1][user_column]]

            if "prompt" in pred_args and pred_args["prompt"] is not None:
                user_prompt = pred_args["prompt"]
            else:
                user_prompt = args["using"].get("prompt", "")

            prompt_template = (
                f"{user_prompt}\n"
                f"---------------------\n"
                f"We have provided context information below. \n"
                f"{{context_str}}\n"
                f"---------------------\n"
                f"This is previous conversation history:\n"
                f"{conversation}\n"
                f"---------------------\n"
                f"Given this information, please answer the question: {{query_str}}"
            )

            engine_kwargs["text_qa_template"] = PromptTemplate(prompt_template)

        else:
            input_column = args["using"].get("input_column", None)

            prompt_template = args["using"].get(
                "prompt_template", args.get("prompt_template", None)
            )
            if prompt_template is not None:
                self.create_validation(args=args)
                engine_kwargs["text_qa_template"] = PromptTemplate(prompt_template)

            if input_column is None:
                raise Exception(
                    f"`input_column` must be provided at model creation time or through USING clause when predicting. Please try again."
                )  # noqa

            if input_column not in df.columns:
                raise Exception(
                    f'Column "{input_column}" not found in input data! Please try again.'
                )

            questions = df[input_column]

        index_path = self.model_storage.folder_get("context")
        storage_context = StorageContext.from_defaults(persist_dir=index_path)
        self._get_service_context()
       
        index = load_index_from_storage(
            storage_context
        )
        query_engine = index.as_query_engine(**engine_kwargs)

        results = []

        for question in questions:
            query_results = query_engine.query(
                question
            )  # TODO: provide extra_info in explain_target col
            results.append(query_results.response)

        result_df = pd.DataFrame(
            {"question": questions, args["target"]: results}
        )  # result_df['answer'].tolist()
        return result_df

    def _get_service_context(self) -> None:
        args = self.model_storage.json_get("args")
        engine_storage = self.engine_storage
        openai_api_key = get_api_key('openai', args["using"], engine_storage, strict=True)
        llm_kwargs = {"api_key": openai_api_key}

        if "temperature" in args["using"]:
            llm_kwargs["temperature"] = args["using"]["temperature"]
        if "model_name" in args["using"]:
            llm_kwargs["model_name"] = args["using"]["model_name"]
        if "max_tokens" in args["using"]:
            llm_kwargs["max_tokens"] = args["using"]["max_tokens"]
        #only way this works is by sending the key through openai

        openai.api_key = openai_api_key
        if Settings.llm is None:
            llm = OpenAI()
            Settings.llm = llm
        if Settings.embed_model is None:
            embed_model = OpenAIEmbedding()
            Settings.embed_model = embed_model
        # TODO: all usual params should be added to Settings

    def _setup_index(self, documents):
        self._get_service_context()
        index = VectorStoreIndex.from_documents(
            documents
        )
        return index



import requests


class NPM:

    def __init__(self, package_name: str):
        resp = requests.get("https://api.npms.io/v2/package/" + package_name)
        if not resp or resp.status_code != 200:
            raise Exception(f"Unable to get package datails: '{package_name}'")
        self.data = resp.json()

    def get_data(self):
        return self.data

    @staticmethod
    def is_connected():
        return True if requests.get("https://api.npms.io/v2/search?q=a&size=1").status_code == 200 else False

    def get_cols_in(self, path, cols):
        curr_root = self.data
        for p in path:
            curr_root = curr_root[p]
        req_cols = {}
        for col in cols:
            req_cols[col] = curr_root[col] if col in curr_root else {}
        return req_cols



def start_http(*args, **kwargs):
    from mindsdb.api.http.start import start
    start(*args, **kwargs)


def start_mysql(*args, **kwargs):
    from mindsdb.api.mysql.start import start
    start(*args, **kwargs)


def start_mongo(*args, **kwargs):
    from mindsdb.api.mongo.start import start
    start(*args, **kwargs)


def start_postgres(*args, **kwargs):
    from mindsdb.api.postgres.start import start
    start(*args, **kwargs)


def start_tasks(*args, **kwargs):
    from mindsdb.interfaces.tasks.task_monitor import start
    start(*args, **kwargs)


def start_ml_task_queue(*args, **kwargs):
    from mindsdb.utilities.ml_task_queue.consumer import start
    start(*args, **kwargs)


def start_scheduler(*args, **kwargs):
    from mindsdb.interfaces.jobs.scheduler import start
    start(*args, **kwargs)


def start_mcp(*args, **kwargs):
    """Start the MCP server"""
    from mindsdb.api.mcp.start import start

    start(*args, **kwargs)
